# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'plot_options.ui'
##
## Created by: Qt User Interface Compiler version 6.5.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QApplication,
    QCheckBox,
    QGridLayout,
    QHeaderView,
    QSizePolicy,
    QTableWidget,
    QTableWidgetItem,
    QWidget,
)


class Ui_Plot_Options(object):
    """ """

    def setupUi(self, Plot_Options):
        """

        Parameters
        ----------
        Plot_Options :


        Returns
        -------

        """
        if not Plot_Options.objectName():
            Plot_Options.setObjectName("Plot_Options")
        Plot_Options.resize(400, 283)
        self.gridLayout = QGridLayout(Plot_Options)
        self.gridLayout.setObjectName("gridLayout")
        self.checkBox_log_y2 = QCheckBox(Plot_Options)
        self.checkBox_log_y2.setObjectName("checkBox_log_y2")

        self.gridLayout.addWidget(self.checkBox_log_y2, 2, 0, 1, 1)

        self.checkBox_use_abs_x = QCheckBox(Plot_Options)
        self.checkBox_use_abs_x.setObjectName("checkBox_use_abs_x")
        self.checkBox_use_abs_x.setEnabled(False)

        self.gridLayout.addWidget(self.checkBox_use_abs_x, 0, 1, 1, 1)

        self.checkBox_log_x = QCheckBox(Plot_Options)
        self.checkBox_log_x.setObjectName("checkBox_log_x")

        self.gridLayout.addWidget(self.checkBox_log_x, 0, 0, 1, 1)

        self.checkBox_use_abs_y = QCheckBox(Plot_Options)
        self.checkBox_use_abs_y.setObjectName("checkBox_use_abs_y")
        self.checkBox_use_abs_y.setEnabled(False)

        self.gridLayout.addWidget(self.checkBox_use_abs_y, 1, 1, 1, 1)

        self.checkBox_log_y = QCheckBox(Plot_Options)
        self.checkBox_log_y.setObjectName("checkBox_log_y")

        self.gridLayout.addWidget(self.checkBox_log_y, 1, 0, 1, 1)

        self.checkBox_use_abs_y2 = QCheckBox(Plot_Options)
        self.checkBox_use_abs_y2.setObjectName("checkBox_use_abs_y2")
        self.checkBox_use_abs_y2.setEnabled(False)

        self.gridLayout.addWidget(self.checkBox_use_abs_y2, 2, 1, 1, 1)

        self.tableWidget = QTableWidget(Plot_Options)
        self.tableWidget.setObjectName("tableWidget")

        self.gridLayout.addWidget(self.tableWidget, 3, 0, 1, 2)

        self.retranslateUi(Plot_Options)

        QMetaObject.connectSlotsByName(Plot_Options)

    # setupUi

    def retranslateUi(self, Plot_Options):
        """

        Parameters
        ----------
        Plot_Options :


        Returns
        -------

        """
        Plot_Options.setWindowTitle(
            QCoreApplication.translate("Plot_Options", "Form", None)
        )
        self.checkBox_log_y2.setText(
            QCoreApplication.translate("Plot_Options", "Y-Axis 2 logarithmic", None)
        )
        self.checkBox_use_abs_x.setText(
            QCoreApplication.translate("Plot_Options", "Use Absolute", None)
        )
        self.checkBox_log_x.setText(
            QCoreApplication.translate("Plot_Options", "X-Axis logarithmic", None)
        )
        self.checkBox_use_abs_y.setText(
            QCoreApplication.translate("Plot_Options", "Use Absolute", None)
        )
        self.checkBox_log_y.setText(
            QCoreApplication.translate("Plot_Options", "Y-Axis logarithmic", None)
        )
        self.checkBox_use_abs_y2.setText(
            QCoreApplication.translate("Plot_Options", "Use Absolute", None)
        )

    # retranslateUi



import sys

from fastapi import FastAPI, HTTPException, Depends, status, Response
from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse, FileResponse, RedirectResponse
from fastapi.security import HTTPBearer, HTTPAuthorizationCredentials
from PySide6.QtCore import QThread, Signal
from nomad_camels.frontpanels.settings_window import hash_api_key
from nomad_camels.utility import load_save_functions, variables_handling
import uvicorn
import sqlite3
import os
import threading
import time
import json
import uuid
from pydantic import BaseModel, Field
import math


def sanitize_dict(d):
    for key, value in d.items():
        if isinstance(value, float):
            if math.isinf(value):
                d[key] = "inf"  # Replace with None or another placeholder
            if math.isnan(value):
                d[key] = "nan"
        elif isinstance(value, dict):
            sanitize_dict(value)
        elif isinstance(value, list):
            for i in range(len(value)):
                if isinstance(value[i], float):
                    if math.isinf(value[i]):
                        value[i] = "inf"  # Replace with None or another placeholder
                    if math.isnan(value[i]):
                        value[i] = "nan"
                elif isinstance(value[i], dict):
                    sanitize_dict(value[i])
    return d


class Variables(BaseModel):
    variables: dict = Field(
        ...,
        json_schema_extra={
            "description": "A dictionary of variables to pass to the protocol",
            "example": {"key1": "value1", "key2": "value2"},
        },
    )


# Define the response model
class ProtocolRunResponse(BaseModel):
    check_protocol_status_here: str = Field(
        ..., description="A URL to check the protocol status"
    )


# Define the validate_api_key function
def validate_api_key(api_key: str) -> bool:
    # Establish the SQLite connection and cursor
    data_base_path = os.path.join(load_save_functions.appdata_path, "CAMELS_API.db")
    conn = sqlite3.connect(data_base_path)
    c = conn.cursor()
    hashed_key = hash_api_key(api_key)
    # Catch exception where the api:keys table does not exist anymore
    try:
        c.execute("SELECT * FROM api_keys WHERE key = ?", (hashed_key,))
    except sqlite3.OperationalError as e:
        if "no such table: api_keys" in str(e):
            conn.close()
            return False
        else:
            raise e # Re-raise the exception if it's not the specific one we're looking for
    result = c.fetchone()
    conn.close()
    return result is not None


def write_protocol_result_path_to_db(api_uuid, message="currently running"):
    # Database setup
    data_base_path = os.path.join(load_save_functions.appdata_path, "CAMELS_API.db")
    conn = sqlite3.connect(data_base_path, check_same_thread=False)
    c = conn.cursor()

    # Create table if it doesn't exist
    c.execute(
        """
        CREATE TABLE IF NOT EXISTS protocol_run_status (
            uuid TEXT PRIMARY KEY,
            status TEXT NOT NULL,
            created_at DATETIME DEFAULT CURRENT_TIMESTAMP
        )
        """
    )
    conn.commit()

    # Insert new entry with UUID and status "currently running"
    c.execute(
        """
        INSERT OR REPLACE INTO protocol_run_status (uuid, status)
        VALUES (?, ?)
        """,
        (api_uuid, message),
    )
    conn.commit()
    # Close the database connection
    conn.close()


def is_valid_path(file_path):
    return os.path.exists(file_path)


def cast_or_create_uuid_str(_uuid: uuid.UUID = None) -> str:
    "Function to cast a given UUID as string or create a new one"
    if _uuid is None:
        return str(uuid.uuid4())
    else:
        return str(_uuid)


# Initialize HTTP Bearer Authentication
security = HTTPBearer()


# Define the dependency for API key validation
async def validate_credentials(credentials: HTTPAuthorizationCredentials = Depends(security)):
    api_key = credentials.credentials
    # The api_key is directly taken from credentials.password
    if api_key and validate_api_key(api_key):
        return True
    raise HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail="Invalid API Key",
        headers={"WWW-Authenticate": "Bearer"},
    )


class FastapiThread(QThread):
    # Define signals for communicating with the main window
    # Signal to start a protocol
    start_protocol_signal = Signal(str, str, object)
    # Signal to send error message to main window, clears the fastapi_thread variable
    port_error_signal = Signal(str)
    # Signal to set the available user names
    set_user_signal = Signal(str)
    # Signal to set the sample in the main window
    set_sample_signal = Signal(str)
    # Signal to set the session name in the main window
    set_session_signal = Signal(str)
    # Signal to queue a protocol
    queue_protocol_signal = Signal(str, str)
    # Signal to remove a protocol from the queue
    remove_queue_protocol_signal = Signal(str)
    # Signal to check the checkbox of the next protocol in the queue
    set_checkbox_signal = Signal(str)
    # Signal to change the variables when adding a protocol to the queue
    queue_protocol_with_variables_signal = Signal(str, dict, int, str)
    # Signal to update the variables of a protocol that is already in the queue
    change_variables_queued_protocol_signal = Signal(str, dict, int)

    def __init__(self, main_window, api_port):
        super().__init__()
        # get the main window so that we have access to all the settings and dictionaries
        self.main_window = main_window
        self.api_port = api_port
        self._stop_event = threading.Event()
        self.app = FastAPI()  # Initialize the FastAPI app
        self.protocol_runs = {}  # Dictionary to store the protocol runs

    def run(self):
        app = self.app
        # Define the API endpoints

        # Get the list of available protocols
        @app.get("/api/v1/protocols")
        async def get_protocols(api_key: str = Depends(validate_credentials)):
            """Get the list of available protocols"""
            return JSONResponse(
                content={"Protocols": list(self.main_window.protocols_dict.keys())}
            )

        # Get the available variables of a protocol
        @app.get("/api/v1/protocols/variables/{protocol_name}")
        async def get_protocol_variables(
            protocol_name: str, api_key: str = Depends(validate_credentials)
        ):
            """Get the available variables of a protocol"""
            return JSONResponse(
                content={
                    "Variables": list(
                        self.main_window.protocols_dict[protocol_name].variables
                    )
                }
            )

        # Get a JSON representation of the protocol
        @app.get("/api/v1/protocols/{protocol_name}/JSON")
        async def get_protocol_json(
            protocol_name: str, api_key: str = Depends(validate_credentials)
        ):
            """Get a JSON representation of the protocol"""
            protocol_class_instance = self.main_window.protocols_dict[protocol_name]
            protocol = load_save_functions.get_save_str(protocol_class_instance)
            cleaned_protocol = sanitize_dict(protocol)
            return JSONResponse(
                content=cleaned_protocol,
            )

        # Get JSON representation of the current CAMELS settings (also called preferences)
        @app.get("/api/v1/settings/JSON")
        async def get_settings_json(api_key: str = Depends(validate_credentials)):
            """Get a JSON representation of the current CAMELS settings"""
            settings = self.main_window.preferences
            cleaned_settings = sanitize_dict(settings)
            return JSONResponse(
                content=cleaned_settings,
            )

        # Run a protocol by name
        @app.get(
            "/api/v1/actions/run/protocols/{protocol_name}",
            response_model=ProtocolRunResponse,
        )
        async def run_protocol(
            protocol_name: str, 
            protocol_uuid: uuid.UUID = None,
            api_key: str = Depends(validate_credentials)
        ):
            """Run a protocol by name"""
            # Generate or use the provided syntax checked UUID as casted string
            protocol_uuid = cast_or_create_uuid_str(_uuid=protocol_uuid)

            write_protocol_result_path_to_db(protocol_uuid)

            self.start_protocol_signal.emit(str(protocol_name), protocol_uuid, None)
            return JSONResponse(
                content={
                    "check protocol status here": f"/api/v1/protocols/results/{protocol_uuid}"
                }
            )

        # Run a protocol by name and pass variables to it (only already defined variables can be passed)
        @app.post(
            "/api/v1/actions/run/protocols/{protocol_name}",
            response_model=ProtocolRunResponse,
        )
        async def run_protocol_with_variables(
            protocol_name: str,
            variables: Variables,
            protocol_uuid: uuid.UUID = None,
            api_key: str = Depends(validate_credentials),
        ):
            """Run a protocol by name and pass variables to it (only already defined variables can be passed)"""
            # Get dictionary from the variables model
            variables = variables.model_dump()
            response = await get_protocol_variables(protocol_name)
            response_content = response.body.decode("utf-8")
            accepted_variables = json.loads(response_content)["Variables"]
            # Check if all variables are in accepted variables
            for key in variables["variables"]:
                if key not in accepted_variables:
                    print(
                        f"Variable {key} is not accepted by the protocol {protocol_name}"
                    )
                    raise HTTPException(
                        status_code=400,
                        detail=f"Variable {key} is not accepted by the protocol {protocol_name}",
                    )

            # Generate or use the provided syntax checked UUID as casted string
            protocol_uuid = cast_or_create_uuid_str(_uuid=protocol_uuid)

            write_protocol_result_path_to_db(protocol_uuid)

            self.start_protocol_signal.emit(
                str(protocol_name), protocol_uuid, variables["variables"]
            )
            return JSONResponse(
                content={
                    "check protocol status here": f"/api/v1/protocols/results/{protocol_uuid}"
                }
            )

        # Get the status of a protocol run by UUID
        @app.get("/api/v1/protocols/results/{protocol_uuid}")
        async def get_protocol_status(
            protocol_uuid: str, api_key: str = Depends(validate_credentials)
        ):
            """Get the status of a protocol run by UUID"""
            # Database setup
            data_base_path = os.path.join(
                load_save_functions.appdata_path, "CAMELS_API.db"
            )
            conn = sqlite3.connect(data_base_path, check_same_thread=False)
            c = conn.cursor()

            # Query the status of the protocol run by UUID
            c.execute(
                """
                SELECT status FROM protocol_run_status WHERE uuid = ?
                """,
                (protocol_uuid,),
            )
            result = c.fetchone()
            # Close the database connection
            conn.close()
            if result:
                return JSONResponse(
                    content={"uuid": protocol_uuid, "status": result[0]}
                )
            else:
                return JSONResponse(
                    content={"error": "UUID not found"}, status_code=404
                )

        # Get the literal hdf5 file of a successful protocol run by UUID
        @app.get("/api/v1/protocols/results/{protocol_uuid}/file")
        async def get_protocol_hdf5(
            protocol_uuid: str, api_key: str = Depends(validate_credentials)
        ):
            """Get the literal hdf5 file of a successful protocol run by UUID"""
            # Database setup
            data_base_path = os.path.join(
                load_save_functions.appdata_path, "CAMELS_API.db"
            )
            conn = sqlite3.connect(data_base_path, check_same_thread=False)
            c = conn.cursor()

            # Query the status of the protocol run by UUID
            c.execute(
                """
                SELECT status FROM protocol_run_status WHERE uuid = ?
                """,
                (protocol_uuid,),
            )
            status = c.fetchone()[0]
            if status == "currently running":
                return JSONResponse(
                    content={
                        "error": "Protocol is currently running. You can access the results once the protocol has finished."
                    },
                    status_code=202,
                )
            if status == "added to queue":
                return JSONResponse(
                    content={
                        "error": "Protocol is in the queue and was not started yet. You can access the results once the protocol has finished."
                    },
                    status_code=202,
                )
            file_path = os.path.abspath(status)
            # Close the database connection
            conn.close()
            if file_path:
                if is_valid_path(file_path):
                    return FileResponse(file_path, filename=os.path.basename(file_path))
                else:
                    return JSONResponse(
                        content={"error": "File does not exists"}, status_code=404
                    )
            else:
                return JSONResponse(
                    content={"error": "UUID not found"}, status_code=404
                )

        # Get the current protocol queue
        @app.get("/api/v1/queue")
        async def get_queue(api_key: str = Depends(validate_credentials)):
            """Get the current protocol queue"""
            protocols = list(
                self.main_window.run_queue_widget.protocol_name_variables.values()
            )
            indexed_protocols = [
                [index] + protocol for index, protocol in enumerate(protocols)
            ]
            return JSONResponse(content=indexed_protocols)

        @app.get("/api/v1/actions/queue/protocols/{protocol_name}")
        async def queue_protocol(
            protocol_name: str, 
            protocol_uuid: uuid.UUID = None,
            api_key: str = Depends(validate_credentials)
        ):
            """Add a protocol to the queue by name"""
            # Generate or use the provided syntax checked UUID as casted string
            protocol_uuid = cast_or_create_uuid_str(_uuid=protocol_uuid)
            write_protocol_result_path_to_db(protocol_uuid, "added to queue")
            try:
                self.queue_protocol_signal.emit(str(protocol_name), protocol_uuid)
            except ValueError as e:
                raise HTTPException(
                    status_code=500,
                    detail=f"Failed to add protocol {protocol_name} to queue as the protocol was not found.\n{e}",
                ) from e
            time.sleep(0.05)
            protocol_values = list(
                self.main_window.run_queue_widget.protocol_name_variables.values()
            )
            if len(protocol_values) > 0:
                # Get the name of the last protocol in the queue
                last_protocol = protocol_values[-1]
                last_protocol_name = last_protocol[0]
                # Check if the provided protocol name matches the last one in the queue
                if protocol_name == last_protocol_name:
                    return JSONResponse(
                        content={
                            "status": "success adding protocol to queue",
                            "uuid": protocol_uuid,
                        }
                    )
                else:
                    raise HTTPException(
                        status_code=500,
                        detail=f"Failed to add protocol {protocol_name} to queue as the protocol was not found.",
                    )
            else:
                # Return a failure response if the queue is empty
                raise HTTPException(
                    status_code=500,
                    detail=f"Failed to add protocol {protocol_name} to queue as the protocol was not found.",
                )

        # Add protocol to queue with variables
        @app.post("/api/v1/actions/queue/protocols/{protocol_name}")
        async def queue_protocol_with_variables(
            protocol_name: str,
            variables: Variables,
            protocol_uuid: uuid.UUID = None,
            api_key: str = Depends(validate_credentials),
        ):
            """Add a protocol to the queue by name and pass variables to it"""
            # Get dictionary from the variables model
            variables = variables.model_dump()
            # Generate or use the provided syntax checked UUID as casted string
            protocol_uuid = cast_or_create_uuid_str(_uuid=protocol_uuid)
            write_protocol_result_path_to_db(protocol_uuid, "added to queue")

            try:
                self.queue_protocol_with_variables_signal.emit(
                    str(protocol_name),
                    variables["variables"],
                    -1,
                    protocol_uuid,
                )
            except ValueError as e:
                raise HTTPException(
                    status_code=500,
                    detail=f"Failed to add protocol {protocol_name} to queue as the protocol was not found.\n{e}",
                ) from e
            time.sleep(0.05)
            protocol_values = list(
                self.main_window.run_queue_widget.protocol_name_variables.values()
            )
            if len(protocol_values) > 0:
                # Get the name of the last protocol in the queue
                last_protocol = protocol_values[-1]
                last_protocol_name = last_protocol[0]
                # Check if the provided protocol name matches the last one in the queue
                if protocol_name == last_protocol_name:
                    return JSONResponse(
                        content={
                            "status": "success adding protocol to queue",
                            "uuid": protocol_uuid,
                        }
                    )
                else:
                    raise HTTPException(
                        status_code=500,
                        detail=f"Failed to add protocol {protocol_name} to queue as the protocol was not found.",
                    )
            else:
                # Return a failure response if the queue is empty
                raise HTTPException(
                    status_code=500,
                    detail=f"Failed to add protocol {protocol_name} to queue as the protocol was not found.",
                )

        # Change variables of a protocol already in the queue
        @app.post("/api/v1/actions/queue/variables/protocols/{protocol_name}_{index}")
        async def change_variables_queued_protocol(
            protocol_name: str,
            index: int,
            variables: Variables,
            api_key: str = Depends(validate_credentials),
        ):
            """Change the variables of a protocol already in the queue"""
            # Get dictionary from the variables model
            variables = variables.model_dump()
            # Get the current queue
            queue_response = await get_queue(api_key)
            # Extract the queue content from the JSON response
            queue_content = queue_response.body.decode("utf-8")
            queue_list = json.loads(queue_content)
            qt_items = list(
                self.main_window.run_queue_widget.protocol_name_variables.keys()
            )
            # make negative indexes work
            if index < 0:
                index = len(qt_items) + index
            for item in queue_list:
                if item[1] == protocol_name and item[0] == index:
                    try:
                        self.change_variables_queued_protocol_signal.emit(
                            str(qt_items[index]),
                            variables["variables"],
                            index,
                        )
                    except ValueError as e:
                        raise HTTPException(
                            status_code=500,
                            detail=f"Failed to update protocol {protocol_name} in queue as the protocol was not found.\n{e}",
                        ) from e
            time.sleep(0.05)
            protocol_values = list(
                self.main_window.run_queue_widget.protocol_name_variables.values()
            )
            if len(protocol_values) > 0:
                # Get the name of the last protocol in the queue
                last_protocol = protocol_values[-1]
                last_protocol_name = last_protocol[0]
                # Check if the provided protocol name matches the last one in the queue
                if protocol_name == last_protocol_name:
                    return JSONResponse(
                        content={"status": "success updating protocol in queue"}
                    )
                else:
                    raise HTTPException(
                        status_code=500,
                        detail=f"Failed to update protocol {protocol_name} in queue as the protocol was not found.",
                    )
            else:
                # Return a failure response if the queue is empty
                raise HTTPException(
                    status_code=500,
                    detail=f"Failed to update protocol {protocol_name} in queue as the protocol was not found.",
                )

        # Remove protocol from queue
        @app.get("/api/v1/actions/queue/remove/protocols/{protocol_name}_{index}")
        async def remove_protocol(
            protocol_name: str, index: int, api_key: str = Depends(validate_credentials)
        ):
            """Remove a protocol from the queue by name"""
            # Get the current queue
            queue_response = await get_queue(api_key)
            # Extract the queue content from the JSON response
            queue_content = queue_response.body.decode("utf-8")
            queue_list = json.loads(queue_content)
            qt_items = list(
                self.main_window.run_queue_widget.protocol_name_variables.keys()
            )
            # Negative index work as well
            if index < 0:
                index = len(qt_items) + index
            for item in queue_list:
                if item[1] == protocol_name and item[0] == index:
                    self.remove_queue_protocol_signal.emit(str(qt_items[index]))
                    return JSONResponse(
                        content={"status": "success removing protocol from queue"}
                    )
            raise HTTPException(
                status_code=500,
                detail=f"Failed to remove protocol {protocol_name} with index {index} from queue as the protocol was not found at that position.",
            )

        # Check the ready checkbox of protocols in the queue
        @app.get("/api/v1/actions/queue/ready/protocols/{protocol_name}_{index}")
        async def protocol_ready(
            protocol_name: str, index: int, api_key: str = Depends(validate_credentials)
        ):
            """Check the ready checkbox of protocols in the queue"""
            # Get the current queue
            queue_response = await get_queue(api_key)
            # Extract the queue content from the JSON response
            queue_content = queue_response.body.decode("utf-8")
            queue_list = json.loads(queue_content)
            qt_items = list(
                self.main_window.run_queue_widget.protocol_name_variables.keys()
            )
            if index < 0:
                index = len(qt_items) + index
            for item in queue_list:
                if item[1] == protocol_name and item[0] == index:
                    self.set_checkbox_signal.emit(str(qt_items[index]))
                    return JSONResponse(content={"status": "success checking protocol"})
            raise HTTPException(
                status_code=500,
                detail=f"Failed to check next protocol {protocol_name} with index {index} from queue as the protocol was not found at that position.",
            )

        @app.get("/api/v1/samples")
        async def get_samples(api_key: str = Depends(validate_credentials)):
            """Get the list of available samples"""
            return JSONResponse(content=self.main_window.sampledata)

        @app.get("/api/v1/actions/set/samples/{sample_name}")
        async def set_samples(
            sample_name: str, api_key: str = Depends(validate_credentials)
        ):
            """Set the active sample"""
            self.set_sample_signal.emit(str(sample_name))
            time.sleep(0.01)
            if self.main_window.active_sample == sample_name:
                return JSONResponse(content={"status": "success setting sample name"})
            else:
                return JSONResponse(content={"status": "failed setting sample name"})

        @app.get("/api/v1/users")
        async def get_users(api_key: str = Depends(validate_credentials)):
            """Get the list of available users"""
            return JSONResponse(content=self.main_window.userdata)

        @app.get("/api/v1/actions/set/users/{user_name}")
        async def set_users(
            user_name: str, api_key: str = Depends(validate_credentials)
        ):
            """Set the active user"""
            self.set_user_signal.emit(str(user_name))
            time.sleep(0.01)
            if self.main_window.active_user == user_name:
                return JSONResponse(content={"status": "success setting user name"})
            else:
                return JSONResponse(content={"status": "failed setting user name"})

        @app.get("/api/v1/session")
        async def get_session(api_key: str = Depends(validate_credentials)):
            """Get the current session name"""
            return JSONResponse(content=self.main_window.lineEdit_session.text())

        @app.get("/api/v1/actions/set/session/{session_name}")
        async def set_session(
            session_name: str, api_key: str = Depends(validate_credentials)
        ):
            """Set the active session name"""
            self.set_session_signal.emit(str(session_name))
            time.sleep(0.01)
            if self.main_window.lineEdit_session.text() == session_name:
                return JSONResponse(content={"status": "success setting session name"})
            else:
                return JSONResponse(content={"status": "failed setting session name"})

        # Root redirects to the API documentation
        @app.get("/")
        async def root():
            return RedirectResponse(url="/docs", status_code=302)

        # Define a favicon endpoint
        @app.get("/favicon.ico")
        async def favicon():
            icon_path = os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                "../graphics/camels_icon_high_res.ico",
            )
            return FileResponse(icon_path)

        @app.get("/logout")
        async def logout():
            """Logout the user"""
            raise HTTPException(status_code=401, detail="Successful Logout completed")

        # Start the uvicorn server
        try:
            self.server_config = uvicorn.Config(
                app,
                host="0.0.0.0",
                port=int(self.api_port),
                log_level="info",
                reload=False,
            )
            self.server = uvicorn.Server(self.server_config)

            # Run the server
            self.server.run()
        except ValueError:  # Catching ValueError for invalid port conversion
            print(f"Invalid port number: {self.api_port}")
            self.port_error_signal.emit("Invalid port number")
        except Exception as e:
            print(f"Error starting server: {e}")
            self.port_error_signal.emit("Failed to start server")

    def stop_server(self):  #  Method to trigger shutdown
        if self.server is not None:
            self.server.should_exit = True
            self._stop_event.set()
            self.quit()
            self.wait()


if __name__ == "__main__":

    class fake_run_queue_widget:
        protocol_name_variables = {
            "key1": ["execute", {"factor": 3, "results": 0}],
            "key2": ["execute2", {"factor": 3, "results": 0}],
            "key3": ["execute3", {"factor": 3, "results": 0}],
        }

    # Create a fake main window for testing
    class fake_main_window:
        def __init__(self):
            self.protocols_dict = {"test_protocol": "test_protocol"}
            self.run_queue_widget = None
            self.sampledata = {"sample1": "sample1"}
            self.userdata = {"user1": "user1"}
            self.lineEdit_session = None
            self.active_sample = None
            self.active_user = None
            self.run_queue_widget = fake_run_queue_widget()

    api_thread = FastapiThread(main_window=fake_main_window(), api_port=12345)
    api_thread.start()
    api_thread.wait()  # Ensure the main thread waits for the server thread to finish



"""
CURRENTLY DEPRECATED MODULE, NOTHING IS USED
This module contains various threads that may be run from the main UI."""

from PySide6.QtCore import QThread

import importlib

from nomad_camels.utility import variables_handling


# class Run_Protocol(QThread):
#     """
#     DEPRECATED
#     Runs the given protocol with a file at the given path."""
#     sig_step = Signal(int)
#     info_step = Signal(str)
#     protocol_done = Signal()
#
#     def __init__(self):
#         super().__init__()
#         self.protocol = None
#         self.path = None
#         self.popen = None
#         self.current_protocol = None
#         self.total_time = np.inf
#         self.counter = 0
#         self.paused = False
#         self.already_run = False
#         self.ready_to_run = False
#
#     def run(self) -> None:
#         """Runs the given `protocol` at `file_path`. If `sig_step` is
#         provided, the stdout will be written there. If `info_step` is
#         provided, it will update the completed-percentage with each starting
#         loopstep.
#
#         Parameters
#         ----------
#
#         Returns
#         -------
#
#         """
#         cmd = ['.desertenv/Scripts/pythonw', '-c', "from IPython import embed; embed()"]
#         flags = 0
#         if os.name == 'nt':
#             flags = subprocess.CREATE_NO_WINDOW
#         self.popen = subprocess.Popen(cmd, stdout=subprocess.PIPE,
#                                       stderr=subprocess.STDOUT,
#                                       stdin=subprocess.PIPE, bufsize=1,
#                                       creationflags=flags)
#         self.write_to_console('%gui qt5')
#         self.write_to_console('import sys')
#         self.write_to_console('import importlib')
#         self.write_to_console('from PySide6.QtWidgets import QApplication')
#         self.write_to_console('app = QApplication(sys.argv)')
#         self.write_to_console('import bluesky_handling.standard_imports')
#         self.ready_to_run = True
#         for line in iter(self.popen.stdout.readline, b''):
#             text = line.decode().rstrip()
#             if text.startswith("starting loop_step "):
#                 self.sig_step.emit(int(self.counter/self.total_time * 100))
#                 self.counter += 1
#             elif text.startswith("protocol finished!"):
#                 self.sig_step.emit(100)
#                 self.protocol_done.emit()
#             else:
#                 self.info_step.emit(text)
#         self.info_step.emit('\n\n\n')
#         self.sig_step.emit(100)
#
#     def run_protocol(self, path, prot_time):
#         """
#
#         Parameters
#         ----------
#         path :
#
#         prot_time :
#
#
#         Returns
#         -------
#
#         """
#         while not self.ready_to_run:
#             time.sleep(0.1)
#         name = os.path.basename(path)[:-3]
#         self.current_protocol = name
#         self.write_to_console(f'spec = importlib.util.spec_from_file_location("{name}", "{path}")')
#         self.write_to_console(f'{name}_mod = importlib.util.module_from_spec(spec)')
#         self.write_to_console(f'sys.modules[spec.name] = {name}_mod')
#         self.write_to_console(f'spec.loader.exec_module({name}_mod)')
#         self.counter = 0
#         self.total_time = prot_time
#         dark = variables_handling.dark_mode
#         theme = variables_handling.preferences['graphic_theme']
#         self.write_to_console(f'dat_{name} = {name}_mod.main(dark={dark}, used_theme="{theme}")')
#         self.already_run = True
#         # self.write_to_console(f'print(dat_{name})')
#
#     def pause(self):
#         """ """
#         # os.kill(self.p.pid, signal.SIGINT)
#         # self.popen.stdin.write('print("test")\n'.encode())
#         # self.popen.stdin.flush()
#         self.popen.send_signal(signal.CTRL_C_EVENT)
#         self.paused = True
#
#     def abort(self):
#         """ """
#         if not self.paused:
#             self.pause()
#         msg = f'{self.current_protocol}_mod.RE.abort()'
#         self.write_to_console(msg)
#         self.popen.terminate()
#         self.terminate()
#
#     def resume(self):
#         """ """
#         msg = f'{self.current_protocol}_mod.RE.resume()'
#         self.write_to_console(msg)
#
#     def write_to_console(self, msg):
#         """
#
#         Parameters
#         ----------
#         msg :
#
#
#         Returns
#         -------
#
#         """
#         if msg == 'exit()':
#             raise Exception('Exiting the shell is not allowed!')
#         if self.popen is not None:
#             self.popen.stdin.write(bytes(f'{msg}\n', 'utf-8'))
#             self.popen.stdin.flush()
#             self.info_step.emit(f'{msg}\n')
#
# class Run_Protocol_test(QThread):
#     """
#     DEPRECATED
#     Runs the given protocol with a file at the given path."""
#     sig_step = Signal(int)
#     info_step = Signal(str)
#     protocol_done = Signal()
#
#     def __init__(self, RE, main_fun):
#         super().__init__()
#         self.protocol = None
#         self.path = None
#         self.popen = None
#         self.current_protocol = None
#         self.total_time = np.inf
#         self.counter = 0
#         self.paused = False
#         self.already_run = False
#         self.ready_to_run = False
#         self.RE = RE
#         self.main_fun = main_fun
#
#     def run(self) -> None:
#         """Runs the given `protocol` at `file_path`. If `sig_step` is
#         provided, the stdout will be written there. If `info_step` is
#         provided, it will update the completed-percentage with each starting
#         loopstep.
#
#         Parameters
#         ----------
#
#         Returns
#         -------
#
#         """
#         # self.counter = 0
#         # self.total_time = prot_time
#         dark = variables_handling.dark_mode
#         theme = variables_handling.preferences['graphic_theme']
#         dat_ = self.main_fun(RE=self.RE, dark=dark, used_theme=theme)
#         # cmd = ['.desertenv/Scripts/pythonw', '-c', "from IPython import embed; embed()"]
#         # self.popen = subprocess.Popen(cmd, stdout=subprocess.PIPE,
#         #                               stderr=subprocess.STDOUT,
#         #                               stdin=subprocess.PIPE, bufsize=1,
#         #                               creationflags=subprocess.CREATE_NO_WINDOW)
#         # self.write_to_console('%gui qt5')
#         # self.write_to_console('import sys')
#         # self.write_to_console('import importlib')
#         # self.write_to_console('from PySide6.QtWidgets import QApplication')
#         # self.write_to_console('app = QApplication(sys.argv)')
#         # self.write_to_console('import bluesky_handling.standard_imports')
#         # self.ready_to_run = True
#         # for line in iter(self.popen.stdout.readline, b''):
#         #     text = line.decode().rstrip()
#         #     if text.startswith("starting loop_step "):
#         #         self.sig_step.emit(int(self.counter/self.total_time * 100))
#         #         self.counter += 1
#         #     elif text.startswith("protocol finished!"):
#         #         self.sig_step.emit(100)
#         #         self.protocol_done.emit()
#         #     else:
#         #         self.info_step.emit(text)
#         # self.info_step.emit('\n\n\n')
#         # self.sig_step.emit(100)
#
#     def run_protocol(self, path, prot_time):
#         """
#
#         Parameters
#         ----------
#         path :
#
#         prot_time :
#
#
#         Returns
#         -------
#
#         """
#         while not self.ready_to_run:
#             time.sleep(0.1)
#         name = os.path.basename(path)[:-3]
#         self.current_protocol = name
#         self.write_to_console(f'spec = importlib.util.spec_from_file_location("{name}", "{path}")')
#         self.write_to_console(f'{name}_mod = importlib.util.module_from_spec(spec)')
#         self.write_to_console(f'sys.modules[spec.name] = {name}_mod')
#         self.write_to_console(f'spec.loader.exec_module({name}_mod)')
#         self.counter = 0
#         self.total_time = prot_time
#         dark = variables_handling.dark_mode
#         theme = variables_handling.preferences['graphic_theme']
#         self.write_to_console(f'dat_{name} = {name}_mod.main(dark={dark}, used_theme="{theme}")')
#         self.already_run = True
#         # self.write_to_console(f'print(dat_{name})')
#
#     def pause(self):
#         """ """
#         # os.kill(self.p.pid, signal.SIGINT)
#         # self.popen.stdin.write('print("test")\n'.encode())
#         # self.popen.stdin.flush()
#         self.popen.send_signal(signal.CTRL_C_EVENT)
#         self.paused = True
#
#     def abort(self):
#         """ """
#         if not self.paused:
#             self.pause()
#         msg = f'{self.current_protocol}_mod.RE.abort()'
#         self.write_to_console(msg)
#         self.popen.terminate()
#         self.terminate()
#
#     def resume(self):
#         """ """
#         msg = f'{self.current_protocol}_mod.RE.resume()'
#         self.write_to_console(msg)
#
#     def write_to_console(self, msg):
#         """
#
#         Parameters
#         ----------
#         msg :
#
#
#         Returns
#         -------
#
#         """
#         if msg == 'exit()':
#             raise Exception('Exiting the shell is not allowed!')
#         if self.popen is not None:
#             self.popen.stdin.write(bytes(f'{msg}\n', 'utf-8'))
#             self.popen.stdin.flush()
#             self.info_step.emit(f'{msg}\n')


class Manual_Device_Thread(QThread):
    """DEPRECATED"""

    def __init__(self, device, ophyd_class):
        super().__init__()
        self.device = ophyd_class(
            f"{device.custom_name}:",
            name=f"manual_{device.custom_name}",
            **device.get_settings(),
        )
        self.device.wait_for_connection()
        self.device.configure(device.get_config())

    def update_config_settings(self, config=None, settings=None):
        """

        Parameters
        ----------
        config :
             (Default value = None)
        settings :
             (Default value = None)

        Returns
        -------

        """
        config = config or {}
        settings = settings or {}
        self.device.configure(config)
        self.device.update_settings(**settings)


class Additional_Imports_Thread(QThread):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.imports = []
        self.catalog = None
        self.catalog_temporary = False

    def run(self) -> None:
        """ """
        from bluesky import RunEngine
        from bluesky.callbacks.best_effort import BestEffortCallback
        import databroker
        from nomad_camels.frontpanels.manage_instruments import ManageInstruments
        from nomad_camels.nomad_integration import nomad_communication
        import pandas as pd
        from nomad_camels.ui_widgets import add_remove_table
        from nomad_camels.nomad_integration import entry_selection
        from nomad_camels.bluesky_handling import make_catalog
        import warnings
        from nomad_camels.frontpanels.settings_window import Settings_Window
        from nomad_camels.manual_controls.get_manual_controls import (
            New_Manual_Control_Dialog,
        )
        from nomad_camels.manual_controls.get_manual_controls import (
            get_control_by_type_name,
        )
        from nomad_camels.frontpanels.protocol_config import Protocol_Config
        from nomad_camels.ui_widgets.path_button_edit import Path_Button_Dialog
        import importlib, bluesky, ophyd, time
        from nomad_camels.nomad_integration import file_uploading
        from nomad_camels.utility import databroker_export, device_handling
        from nomad_camels.bluesky_handling import protocol_builder
        from nomad_camels.tools import device_driver_builder
        from nomad_camels.tools import EPICS_driver_builder
        from nomad_camels.tools import databroker_exporter
        from nomad_camels.bluesky_handling import helper_functions

        self.imports.append(RunEngine)
        self.imports.append(BestEffortCallback)
        self.imports.append(databroker)
        self.imports.append(ManageInstruments)
        self.imports.append(nomad_communication)
        self.imports.append(pd)
        self.imports.append(add_remove_table)
        self.imports.append(entry_selection)
        self.imports.append(make_catalog)
        self.imports.append(warnings)
        self.imports.append(Settings_Window)
        self.imports.append(New_Manual_Control_Dialog)
        self.imports.append(get_control_by_type_name)
        self.imports.append(Protocol_Config)
        self.imports.append(Path_Button_Dialog)
        self.imports.append(importlib)
        self.imports.append(bluesky)
        self.imports.append(ophyd)
        self.imports.append(time)
        self.imports.append(device_handling)
        self.imports.append(file_uploading)
        self.imports.append(databroker_export)
        self.imports.append(protocol_builder)
        self.imports.append(device_driver_builder)
        self.imports.append(EPICS_driver_builder)
        self.imports.append(databroker_exporter)
        self.imports.append(helper_functions)

        try:
            self.catalog = databroker.catalog[
                variables_handling.preferences["databroker_catalog_name"]
            ]
        except KeyError:
            print("Could not find databroker catalog, using temporary")
            self.catalog = databroker.temp().v2
            self.catalog_temporary = True



"""This module helps to synchronize information between different modules. To
this aim, it holds several variables that may be read from different places.
Furthermore, some functions to work with those variables are provided.

Attributes
----------
    preset : str
        The name of the currently used preset of CAMELS.
    device_driver_path : str, path
        The path, where to find local drivers.
    meas_files_path : str, path
        The path, where to write the measurement files, i.e. data.
    CAMELS_path : str, path
        The path to the current installation of CAMELS.
    preferences : dict
        The currently used preferences.
    protocols : dict{"<protocol_name>": protocol}
        The available protocols.
    protocol_variables : dict{"<name>": <value>}
        The variables provided by the currently viewed protocol.
    channels : dict{"<name>": channel}
        All available channels provided by the configured instruments.
    loop_step_variables : dict{"<name>": <value>}
        The variables provided by the steps of the currently viewed protocol.
    devices : dict
        All configured instruments/devices.
    current_protocol = Measurement_Protocol
        The protocol, that is currently being used.
    dark_mode : bool
        Whether dark-mode is currently active.
    copied_step : Loop_Step
        The last step, that was copied.
    read_channel_sets : list[set]
        Sets of the different channel-compositions for read-channels. Used to distinguish different reads with different channels for bluesky
    read_channel_names : list[str]
        Names of the different read-channel steps in use. Used to distinguish the different reads.
    evaluation_functions_names : dict
        Used to provide the right-click menu to use mathematical functions.
    operator_names : dict
        Used to provide the right-click menu to use mathematical operators.
"""

from ast import literal_eval, parse
from PySide6.QtWidgets import QMenu
from PySide6.QtGui import QColor, QAction

import numpy as np
from nomad_camels.bluesky_handling import evaluation_helper
from nomad_camels.ui_widgets.warn_popup import WarnPopup


preset = ""
device_driver_path = ""
meas_files_path = ""
CAMELS_path = ""

preferences = {}

protocols = {}
protocol_variables = {}
channels = {}
config_channels = {}
loop_step_variables = {}
devices = {}
watchdogs = {}
current_protocol = None
dark_mode = False

copied_step = None

read_channel_sets = []
read_channel_names = []

evaluation_functions_names = {
    "randint()": "randint(x) - random integer below x",
    "rand()": "rand() - random float between 0 and 1",
    "round()": "round(x) - round number to nearest integer",
    "exp()": "exp(x) - exponential function of x",
    "sqrt()": "sqrt(x) - square root of x",
    "log()": "ln(x) - natural logarithm of x",
    "sin()": "sin(x) - sine of x",
    "cos()": "cos(x) - cosine of x",
    "tan()": "tan(x) - tangent of x",
    "sinh()": "sinh(x) - hyperbolic sine of x",
    "cosh()": "cosh(x) - hyperbolic cosine of x",
    "tanh()": "tanh(x) - hyperbolic tangent of x",
    "arctan()": "arctan(x) - arcus tangent of x",
    "arcsin()": "arcsin(x) - arcus sine of x",
    "arccos()": "arccos(x) - arcus cosine of x",
    "arcsinh()": "arcsinh(x) - area hyperbolic sine of x",
    "arccosh()": "arccosh(x) - area hyperbolic cosine of x",
    "arctanh()": "arctanh(x) - area hyperbolic tangent of x",
    "sinc()": "sinc(x) - sinc function of x",
}

operator_names = {
    "+": "add",
    "-": "subtract",
    "/": "divide",
    "*": "multiply",
    "**": "to the power of",
    "%": "modulus",
    "==": "equals",
    "<": "less than",
    ">": "greater than",
    "<=": "less or equal",
    ">=": "greater or equal",
    "and": "logical AND",
    "or": "logical OR",
    "not": "logical negation",
}


def open_link(link):
    """
    Open a link in the default program. Should work in all operating systems.

    Parameters
    ----------
    link : str
        The link to open.
    """
    import os
    import platform
    import subprocess

    if platform.system() == "Windows":
        os.startfile(link)
    else:
        opener = "open" if platform.system() == "Darwin" else "xdg-open"
        subprocess.call([opener, link])


def get_non_channel_functions():
    functions = []
    for device in devices.values():
        functions += device.get_non_channel_functions()
    return functions


def get_output_channels():
    """Goes through all channels and returns a list of the names of those, that
    are outputs."""
    outputs = []
    for channel in channels:
        if channels[channel].output:
            outputs.append(channel)
    return outputs


def get_color(color="", string=False):
    """Returns the respective QColor or rgb-code(if `string`) for
    `color`, taking dark-mode into account.

    Parameters
    ----------
    color : str
        (Default value = '')
        The name of the color asked for. Possible values are "red" / "r",
        "strong_red", "green" / "g", "dark_green", "grey" / "gray", "blue" / "b"
        "black" (white if `dark_mode`), "orange". Otherwise white (or black, if
        `dark_mode`) is returned.
    string : bool
        (Default value = False)
        If True, only the string of the rgb will be returned, not the QColor.
    """
    if color == "red" or color == "r":
        rgb = (255, 180, 180)
        if dark_mode:
            rgb = (75, 0, 0)
    elif color == "strong_red":
        rgb = (230, 0, 0)
    elif color == "green" or color == "g":
        rgb = (180, 255, 180)
        if dark_mode:
            rgb = (0, 75, 0)
    elif color == "dark_green":
        rgb = (32, 175, 32)
    elif color == "grey" or color == "gray":
        rgb = (169, 169, 169)
    elif color == "blue" or color == "b":
        rgb = (180, 180, 255)
        if dark_mode:
            rgb = (0, 0, 75)
    elif color == "black":
        rgb = (0, 0, 0)
        if dark_mode:
            rgb = (255, 255, 255)
    elif color == "orange":
        rgb = (255, 170, 25)
    else:
        rgb = (255, 255, 255)
        if dark_mode:
            rgb = (0, 0, 0)
    if string:
        return str(rgb)
    return QColor(*rgb)


def get_menus(connect_function, pretext="Insert"):
    """Providing QMenus with the `connect_function` for each action,
    containing all the variables, channels, functions and operators.

    Parameters
    ----------
    connect_function : callable
        The function that should be executed when the action is clicked. It gets
        the variable's name as a value.
    pretext : str
        (Default value = 'Insert')
        This string will be written in front of the menus.

    Returns
    -------
    menus : list[QMenu]
        the created menus
    actions : list[list[QAction]]
        lists of the individual actions in the menus
    """
    variable_menu = QMenu(f"{pretext} Variable")
    channel_menu = QMenu(f"{pretext} Channel-Value")
    function_menu = QMenu(f"{pretext} Function")
    operator_menu = QMenu(f"{pretext} Operator")
    channel_actions = []
    operator_actions = []
    actions = []
    function_actions = []
    add_actions_from_dict(channels, channel_actions, connect_function)
    add_actions_from_dict(protocol_variables, actions, connect_function)
    add_actions_from_dict(loop_step_variables, actions, connect_function)
    add_actions_from_dict({"StartTime": 1, "ElapsedTime": 1}, actions, connect_function)
    add_actions_from_dict(operator_names, operator_actions, connect_function)
    add_actions_from_dict(
        evaluation_functions_names, function_actions, connect_function
    )
    channel_menu.addActions(channel_actions)
    variable_menu.addActions(actions)
    operator_menu.addActions(operator_actions)
    function_menu.addActions(function_actions)
    if pretext == "Insert":
        menus = [channel_menu, variable_menu, function_menu]
        actions = [channel_actions, actions, function_actions]
    else:
        menus = [channel_menu, variable_menu, operator_menu, function_menu]
        actions = [channel_actions, actions, operator_actions, function_actions]
    return menus, actions


def add_actions_from_dict(dictionary, actions, connect_function, add_string=""):
    """
    The values of `dictionary` are handed to the `connect_function` when
    clicking on the respective action named with the keys of the dictionary. The
    created actions are added to `actions`.

    Parameters
    ----------
    dictionary : dict
        the keys become the names of the actions, the values are handed to the
        `connect_function`
    actions : list
        the created actions will be added to this list
    connect_function : callable
        this function is called, when one of the actions is clicked
    add_string : str
        (Default value = '')
        added in front of the values of the dictionary for the connect_function
    """
    for var in sorted(dictionary, key=lambda x: x.lower()):
        if isinstance(dictionary[var], dict):
            add_actions_from_dict(dictionary[var], actions, connect_function, f"{var}:")
        else:
            addvar = f"{add_string}{var}"
            action = QAction(addvar)
            action.triggered.connect(lambda state=None, x=addvar: connect_function(x))
            actions.append(action)


def check_eval(s):
    """Checks, whether the string `s` can be evaluated. Returns True if it is
    possible, otherwise False.

    Parameters
    ----------
    s : str
        the string that should be checked
    """
    try:
        namespace = dict(evaluation_helper.base_namespace)
        namespace.update(protocol_variables)
        namespace.update(loop_step_variables)
        for channel in channels:
            namespace.update({channel: 1})
        # utils.call_or_eval_one(s, namespace)
        evaluation_helper.get_eval(s, namespace)
        return True
    except Exception:
        return False


def get_eval(s):
    """
    Evaluates the string `s` with the namespace of `protocol_variables` and
    `loop_step_variables` in addition to `evaluation_helper.base_namespace`.
    Returns the evaluated value.

    Parameters
    ----------
    s : str
        the string that should be evaluated
    """
    try:
        namespace = dict(evaluation_helper.base_namespace)
        namespace.update(protocol_variables)
        namespace.update(loop_step_variables)
        for channel in channels:
            namespace.update({channel: 1})
        return evaluation_helper.get_eval(s, namespace)
    except:
        return np.nan


def get_data(s):
    """Used instead of `get_eval` when there is no specific namespace. It simply
    uses `ast.literal_eval` and if it does not work, the string `s` is returned.

    Parameters
    ----------
    s : str
        the string to be evaluated
    """
    if not s:
        return ""
    try:
        lit = literal_eval(s)
    except ValueError:
        return s
    except SyntaxError:
        return s
    return lit


def check_data_type(s):
    """Returns the datatype of the string-evaluation of s.

    Parameters
    ----------
    s : str
        the string that should be checked
    """
    if not isinstance(s, str):
        return str(type(s))
    if not s:
        return ""
    try:
        lit = literal_eval(s)
    except ValueError:
        return "String"
    except SyntaxError:
        return "String"
    return str(type(lit))


def get_write_from_data_type(s):
    """Used for writing longer strings. Since the strings should stay strings in
    the written files, if the evaluated datatype of `s` is str, quotation marks
    will be added around the value, otherwise nothing is done."""
    t = check_data_type(s)
    if t == "String":
        return f'"{s}"'
    return s


def check_variable_name(name, raise_not_warn=False, parent=None):
    """Checks whether `name` is a valid name for a variable. It checks, whether
    the name mirrors a builtin function or keyword. If not, it is tried with ast
    to parse the statement `<name> = None`, i.e. whether it works as a variable
    name. If anything fails, either an exception is raised (if `raise_not_warn`)
    or a `WarnPopup` is called.

    Parameters
    ----------
    name : str
        the name that should be checked
    raise_not_warn : bool
        (Default value = False)
        if True, an exception is raised if a check fails, otherwise a WarnPopup
    parent : QWidget
        (Default value = None)
        the parent widget for a possibly called WarnPopup
    """
    try:
        built_check = name in vars(__builtins__)
    except:
        built_check = name in __builtins__
    if built_check:
        text = (
            f'The name "{name}" is a python builtin function! Please use another name.'
        )
        if raise_not_warn:
            raise Exception(text)
        WarnPopup(parent, text, "Invalid Name")
        return False
    try:
        parse(f"{name} = None")
    except (ValueError, SyntaxError, TypeError) as e:
        text = f'The name "{name}" is not a valid name! Remove e.g. spaces or special characters.'
        if raise_not_warn:
            raise Exception(text) from e
        WarnPopup(parent, text, "Invalid Name")
        return False
    return True



# THIS IS DEPRECATED

print(
    "You are using a deprecated version of the code. Please update to the newest version."
)

import threading
import numpy as np

import matplotlib

matplotlib.use("Qt5Agg")
import matplotlib.pyplot as plt
from bluesky.callbacks.mpl_plotting import LiveScatter
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT

from PySide6.QtWidgets import QWidget, QGridLayout, QPushButton
from PySide6.QtCore import Signal, QObject
from PySide6.QtGui import QIcon

from nomad_camels.bluesky_handling.evaluation_helper import Evaluator

from nomad_camels.main_classes.plot_widget import MPLwidget

from nomad_camels.utility.plot_placement import place_widget
from importlib import resources
from nomad_camels import graphics

stdCols = plt.rcParams["axes.prop_cycle"].by_key()["color"]


class PlotWidget_2D(QWidget):
    """ """

    closing = Signal()

    def __init__(
        self,
        x_name,
        y_name,
        z_name,
        *,
        xlim=None,
        ylim=None,
        zlim=None,
        parent=None,
        namespace=None,
        zlabel="",
        ylabel="",
        xlabel="",
        title="",
        stream_name="primary",
        **kwargs,
    ):
        super().__init__(parent=parent)
        canvas = MPLwidget()
        self.ax = canvas.axes
        self.stream_name = stream_name
        self.x_name = x_name
        self.y_name = y_name
        self.z_name = z_name
        eva = Evaluator(namespace=namespace)
        self.livePlot = LivePlot_2D(
            x_name,
            y_name,
            z_name,
            xlim=xlim,
            ylim=ylim,
            zlim=zlim,
            ax=self.ax,
            xlabel=xlabel,
            ylabel=ylabel,
            zlabel=zlabel,
            cmap="viridis",
            evaluator=eva,
            stream_name=stream_name,
            **kwargs,
        )
        self.livePlot.new_data.connect(self.show)
        self.toolbar = NavigationToolbar2QT(canvas, self)

        self.pushButton_autoscale = QPushButton("Autoscale")
        self.pushButton_autoscale.clicked.connect(self.autoscale)

        layout = QGridLayout()
        layout.addWidget(canvas, 0, 1, 1, 3)
        layout.addWidget(self.toolbar, 1, 2, 1, 2)
        layout.addWidget(self.pushButton_autoscale, 1, 1)
        self.setLayout(layout)

        self.setWindowTitle(title or f"{z_name} 2D")
        self.setWindowIcon(QIcon(str(resources.files(graphics) / "camels_icon.png")))
        place_widget(self)

    def autoscale(self):
        """ """
        self.ax.autoscale()
        self.ax.figure.canvas.draw_idle()

    def clear_plot(self):
        """ """
        self.livePlot.clear_plot()

    def closeEvent(self, a0):
        """

        Parameters
        ----------
        a0 :


        Returns
        -------

        """
        self.closing.emit()
        super().closeEvent(a0)


class LivePlot_2D(LiveScatter, QObject):
    """ """

    new_data = Signal()

    def __init__(
        self,
        x,
        y,
        z,
        *,
        xlim=None,
        ylim=None,
        zlim=None,
        ax=None,
        xlabel="",
        ylabel="",
        zlabel="",
        cmap="viridis",
        evaluator=None,
        stream_name="primary",
        **kwargs,
    ):
        LiveScatter.__init__(
            self, x, y, z, xlim=xlim, ylim=ylim, clim=zlim, ax=ax, **kwargs
        )
        QObject.__init__(self)
        self.__setup_lock = threading.Lock()
        self.__setup_event = threading.Event()
        self._minx, self._maxx, self._miny, self._maxy = (None,) * 4
        self._xdata, self._ydata, self._Idata = [], [], []

        def setup():
            """ """
            # Run this code in start() so that it runs on the correct thread.
            nonlocal x, y, z, xlim, ylim, zlim, cmap, ax, kwargs  # noqa: E741
            with self.__setup_lock:
                if self.__setup_event.is_set():
                    return
                self.__setup_event.set()
            import matplotlib.colors as mcolors

            if ax is None:
                fig, ax = plt.subplots()
                fig.show()
            ax.cla()
            self.x = x
            self.y = y
            self.I = z  # noqa: E741
            ax.set_xlabel(xlabel or x)
            ax.set_ylabel(ylabel or y)
            ax.set_aspect("equal")
            self._sc = []
            self.ax = ax
            ax.margins(0.1)
            self._xdata, self._ydata, self._Idata = [], [], []
            self._norm = mcolors.Normalize()

            self.xlim = xlim
            self.ylim = ylim
            if xlim is not None:
                ax.set_xlim(xlim)
            if ylim is not None:
                ax.set_ylim(ylim)
            if zlim is not None:
                self._norm.vmin, self._norm.vmax = zlim
            self.clim = zlim
            self.cmap = cmap
            self.kwargs = kwargs
            self.kwargs.setdefault("edgecolor", "face")
            self.kwargs.setdefault("s", 50)

        self.__setup = setup
        self.zlabel = zlabel

        self.stream_name = stream_name
        self.eva = evaluator
        self.sc = None
        self.cb = None
        self.pcolormesh = None
        self.desc = ""

    def start(self, doc):
        """

        Parameters
        ----------
        doc :


        Returns
        -------

        """
        self.__setup()
        self._xdata.clear()
        self._ydata.clear()
        self._Idata.clear()
        sc = self.ax.scatter(
            self._xdata,
            self._ydata,
            c=self._Idata,
            norm=self._norm,
            cmap=self.cmap,
            marker=",",
            **self.kwargs,
        )
        self.ax.set_aspect("auto")
        self._sc.append(sc)
        self.sc = sc
        self.cb = self.ax.figure.colorbar(sc, ax=self.ax)
        self.cb.set_label(self.zlabel or self.I)

    def descriptor(self, doc):
        """

        Parameters
        ----------
        doc :


        Returns
        -------

        """
        if doc["name"] == self.stream_name:
            self.desc = doc["uid"]

    def event(self, doc):
        """

        Parameters
        ----------
        doc :


        Returns
        -------

        """
        if doc["descriptor"] != self.desc:
            return
        try:
            x = doc["data"][self.x]
        except KeyError:
            if self.x in ("time", "seq_num"):
                x = doc[self.x]
            else:
                if not self.eva.is_to_date(doc["time"]):
                    self.eva.event(doc)
                x = self.eva.eval(self.x)
        try:
            y = doc["data"][self.y]
        except KeyError:
            if not self.eva.is_to_date(doc["time"]):
                self.eva.event(doc)
            y = self.eva.eval(self.y)
        try:
            I = doc["data"][self.I]
        except KeyError:
            if not self.eva.is_to_date(doc["time"]):
                self.eva.event(doc)
            I = self.eva.eval(self.I)
        self.update(x, y, I)
        self.new_data.emit(None)

    def make_colormesh(self, x_shape=None, y_shape=None):
        if x_shape is None and y_shape is None:
            return None
            if len(self._xdata) != len(self._ydata) or len(self._xdata) != len(
                self._Idata
            ):
                return None
            try:
                x, y = np.meshgrid(self._xdata, self._ydata)
                c = np.tile(self._Idata, (len(self._Idata), 1))
                return x, y, c
            except Exception as e:
                print(e)
                return None
        elif x_shape is not None and y_shape is None:
            y_shape = int(np.array(self._xdata).size / x_shape)
        elif x_shape is None and y_shape is not None:
            x_shape = int(np.array(self._ydata).size / y_shape)
        try:
            x = np.array(self._xdata).reshape((x_shape, y_shape))
            y = np.array(self._ydata).reshape((x_shape, y_shape))
            c = np.array(self._Idata).reshape((x_shape, y_shape))
            return x, y, c
        except Exception as e:
            print(e)
            return None

    def update(self, x, y, I):
        x_shape = None
        y_shape = None
        if np.isscalar(I):
            I = np.array([I])
        if np.isscalar(x):
            x = np.full(I.shape, x)
        else:
            x_shape = x.size
        if np.isscalar(y):
            y = np.full(I.shape, y)
        else:
            y_shape = y.size

        # if one is None all are
        if self._minx is None:
            self._minx = np.min(x)
            self._maxx = np.max(x)
            self._miny = np.min(y)
            self._maxy = np.max(y)

        self._xdata.extend(x)
        self._ydata.extend(y)
        self._Idata.extend(I)
        mesh = self.make_colormesh(x_shape, y_shape)
        no_mesh = True
        if mesh:
            try:
                x, y, c = mesh
                if self.pcolormesh:
                    self.pcolormesh.remove()
                self.pcolormesh = self.ax.pcolormesh(x, y, c)
                self.pcolormesh.set_visible(True)
                self.sc.set_visible(False)
                no_mesh = False
            except:
                no_mesh = True
        if no_mesh:
            offsets = np.vstack([self._xdata, self._ydata]).T
            self.sc.set_offsets(offsets)
            self.sc.set_array(np.asarray(self._Idata))
            if self.pcolormesh:
                self.pcolormesh.remove()
            self.sc.set_visible(True)

        if self.xlim is None:
            self._minx, self._maxx = np.minimum(np.min(x), self._minx), np.maximum(
                np.max(x), self._maxx
            )
            self.ax.set_xlim(self._minx, self._maxx)

        if self.ylim is None:
            self._miny, self._maxy = np.minimum(np.min(y), self._miny), np.maximum(
                np.max(y), self._maxy
            )
            self.ax.set_ylim(self._miny, self._maxy)

        if self.clim is None:
            clim = np.nanmin(self._Idata), np.nanmax(self._Idata)
            self.sc.set_clim(*clim)
        self.ax.figure.canvas.draw_idle()

    def clear_plot(self):
        """ """
        self._xdata.clear()
        self._ydata.clear()
        self._Idata.clear()



from PySide6.QtWidgets import (
    QWidget,
    QGridLayout,
    QPushButton,
    QTableView,
    QLabel,
    QComboBox,
    QMenu,
    QDialog,
    QDialogButtonBox,
    QMessageBox,
    QLineEdit,
    QVBoxLayout,
    QWidgetAction,
)
from PySide6.QtGui import QStandardItemModel, QStandardItem, QFont, QBrush, QKeyEvent
from PySide6.QtCore import Qt, Signal

import numpy as np
import pandas as pd

from nomad_camels.utility import variables_handling


class AddRemoveTable(QWidget):
    """This widget provides a QTableView and two buttons for adding /
    removing rows / columns.

    Parameters
    ----------

    Returns
    -------

    """

    sizechange = Signal()
    added = Signal(int)
    removed = Signal(int)

    def __init__(
        self,
        addLabel="+",
        removeLabel="-",
        horizontal=True,
        editables=None,
        checkables=(),
        headerLabels=None,
        orderBy=None,
        parent=None,
        tableData=None,
        title="",
        comboBoxes=None,
        subtables=None,
        growsize=False,
        checkstrings=None,
        askdelete=False,
        fixedsize=False,
        enableds=None,
        default_values=None,
    ):
        """

        Parameters
        ----------
        addLabel : str, default "+"
            The label of the add-button
        removeLabel : str, default "-"
            The label of the remove-button
        horizontal : bool, default True
            True if the header is horizontal (thus the data in columns),
            False if it is the other way round
        editables : int, list of int, default None
            Positions of the columns (if horizontal) that are editable,
            If None, all are
        checkables : int, list of int, default None
            Positions of the columns (if horizontal) that have a
            Checkbox, if None, none have
        headerLabels : list of str, default None
            The labels of the header, they are also used to provide the
            format of the data returned by the AddRemoveTable
        orderBy : None
            not used (yet)
        parent : QWidget
            Parent of the Widget
        tableData : dict, pd.DataFrame, list, default None
            The data that should be put into the table. Should have the
            format of headerLabels. If it is a dict, the format should
            best look like {"column1": [data1, data2, ...], ...}
            A list should be used, if the table is a subtable inside
            another one
        title : str, default ""
            adds a QLabel on the top-right with `title`. If an empty
            string, no label is added
        comboBoxes : dict of lists of str
            the keys of the dict specify in which columns (same name), a
            comboBox should be used, the corresponding value gives the
            possible choices of the comboBox
        subtables : dict of lists
            the keys of the dict specify in which columns (same name),
            an additional AddRemoveTable should be contained
        growsize : bool, default True
            Whether to increase the MaximumSize of the Widget when more
            data is contained
        checkstrings : int, list of int, default None
            Positions of the columns (if horizontal) that are to be
            checked for validity, if None, none are
        askdelete : bool, default False
            Whether to have a PopUp make sure, that the user wants to
            delete the selected column
        fixedsize : bool, default False
            Whether the widget should have a fixed size
        enableds : int, list of int, default None
            Positions of the columns (if horizontal) that are enabled,
            If None, all are
        default_values : dict, default None
            A dict with the default values for the table. The keys are
            the column names and the values are the default values. If
            a column is not in the dict, the default value is an empty
            string. If None, all default values are empty strings.
            The default values are used for new entries.
        """
        super().__init__(parent)
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.context_menu)
        layout = QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        self.askdelete = askdelete
        self.checkables = checkables if not type(checkables) is int else [checkables]
        if checkstrings is None:
            checkstrings = []
        self.checkstrings = (
            checkstrings if not type(checkstrings) is int else [checkstrings]
        )
        self.horizontal = horizontal
        self.orderBy = orderBy
        self.comboBoxes = {} if comboBoxes is None else comboBoxes
        self.subtables = {} if subtables is None else subtables
        self.default_values = {} if default_values is None else default_values
        self.growsize = growsize
        self.fixedsize = fixedsize
        self.boxes = []
        self.tables = []
        if headerLabels is None and tableData is not None:
            headerLabels = tableData.keys()
        elif tableData is None and headerLabels is not None:
            tableData = {}
        elif tableData is None and headerLabels is None:
            raise Exception(
                "Cannot create a table without Data Format (AddRemoveTable)"
            )
        if isinstance(tableData, dict):
            tableData = pd.DataFrame(tableData)
        for label in headerLabels:
            if label not in tableData:
                tableData.insert(
                    headerLabels.index(label), label, [""] * len(tableData)
                )
        self.tableData = tableData
        if editables is None:
            editables = range(len(headerLabels))
        self.editables = editables if not isinstance(editables, int) else [editables]
        if enableds is None:
            enableds = range(len(headerLabels))
        self.enableds = enableds if not isinstance(enableds, int) else [enableds]

        self.addButton = QPushButton(addLabel)
        self.addButton.setMaximumHeight(24)
        self.addButton.setStyleSheet(
            "QPushButton {\n"
            "                                background-color: #4CAF50; \n"
            "                                color: white; \n"
            "                                border: none; \n"
            "                                padding: 0px; \n"
            "                                padding-bottom: 5px;\n"
            "                                text-align: center; \n"
            "                                text-decoration: none; \n"
            "                                font-size: 15px; \n"
            "                                margin: 2px 2px; \n"
            "                                border-radius: 6px;\n"
            "								font-weight: bold;\n"
            "                            }\n"
            "\n"
            "                            QPushButton:hover {\n"
            "                                background-color: #45a049;\n"
            "                            }\n"
            "\n"
            "                            QPushButton:disabled {\n"
            "                                background-color: #808080;\n"
            "                                color: white;\n"
            "                            }"
        )
        self.removeButton = QPushButton(removeLabel)
        self.removeButton.setMaximumHeight(24)
        self.removeButton.setStyleSheet(
            "QPushButton {\n"
            "                                background-color:  #FF3333; \n"
            "                                color: white; \n"
            "                                border: none; \n"
            "                                padding: 0px; \n"
            "                                padding-bottom: 5px;\n"
            "                                text-align: center; \n"
            "                                text-decoration: none; \n"
            "                                font-size: 15px; \n"
            "                                margin: 2px 2px; \n"
            "                                border-radius: 6px;\n"
            "								font-weight: bold;\n"
            "                            }\n"
            "\n"
            "            QPushButton:hover {\n"
            "                background-color: #B22222;\n"
            "            }"
            "\n"
            "                            QPushButton:disabled {\n"
            "                                background-color: #808080;\n"
            "                                color: white;\n"
            "                            }"
        )
        self.table = QTableView()
        self.table_model = QStandardItemModel()
        self.table.setModel(self.table_model)
        mover = 0
        if len(title):
            label = QLabel(title)
            layout.addWidget(label, 0, 0)
            font = QFont()
            font.setBold(True)
            label.setStyleSheet("font-size: 9pt")
            label.setFont(font)
            mover = 1
        layout.addWidget(self.addButton, 0, 0 + mover)
        layout.addWidget(self.removeButton, 0, 1 + mover)
        layout.addWidget(self.table, 1, 0, 1, 2 + mover)

        self.headerLabels = headerLabels
        if horizontal:
            self.table.verticalHeader().hide()
            if len(headerLabels):
                self.table_model.setHorizontalHeaderLabels(headerLabels)
                self.table_model.setColumnCount(len(headerLabels))
            else:
                self.table_model.setColumnCount(1)
                self.table.horizontalHeader().hide()
        else:
            self.table.horizontalHeader().hide()
            if len(headerLabels):
                self.table_model.setVerticalHeaderLabels(headerLabels)
                self.table_model.setRowCount(len(headerLabels))
            else:
                self.table_model.setRowCount(1)
                self.table.verticalHeader().hide()

        self.addButton.clicked.connect(lambda x: self.add())
        self.removeButton.clicked.connect(self.remove)
        self.table.clicked.connect(self.table.resizeColumnsToContents)
        self.table.clicked.connect(self.table.resizeRowsToContents)
        self.load_table_data()
        self.table.resizeColumnsToContents()
        self.table.resizeRowsToContents()
        self.update_max_hight()
        self.table_model.itemChanged.connect(self.check_string)
        self.table.selectionModel().selectionChanged.connect(self.check_selection)
        self.check_selection()

    def check_selection(self):
        if self.table.selectedIndexes():
            self.removeButton.setEnabled(True)
        else:
            self.removeButton.setEnabled(False)

    def update_max_hight(self):
        """ """
        if self.growsize:
            self.setMaximumHeight(90 + self.table_model.rowCount() * 100)
            self.sizechange.emit()
        elif self.fixedsize:
            self.setMaximumHeight(100)
            self.setMaximumWidth(200)
        elif not self.horizontal:
            self.setMaximumHeight(30 * len(self.headerLabels) + 30)

    def change_table_data(self, tableData, change_focus=False):
        """

        Parameters
        ----------
        tableData :


        Returns
        -------

        """
        if isinstance(tableData, dict):
            tableData = pd.DataFrame(tableData)
        self.tableData = tableData
        self.load_table_data(change_focus)

    def load_table_data(self, change_focus=False):
        """Putting the `tableData` into the table."""
        if self.horizontal:
            while self.table_model.rowCount():
                self.table_model.removeRow(0)
        else:
            while self.table_model.columnCount():
                self.table_model.removeColumn(0)
        data = np.array(self.tableData)
        for dat in data:
            self.add(dat, change_focus)

    def context_menu(self, pos):
        """Generates the right-click-menu.
        There are entries for inserting (replace) and appending the
        variables, channels, functions and operators.

        Parameters
        ----------
        pos :


        Returns
        -------

        """
        menu = QMenu()
        search_bar = QLineEdit(menu)
        search_bar.setPlaceholderText("Search...")
        # Setting up layout for the menu
        layout = QVBoxLayout()
        layout.addWidget(search_bar)
        container = QWidget()
        container.setLayout(layout)

        # Adding the container widget to the menu
        action = QWidgetAction(menu)
        action.setDefaultWidget(container)
        menu.addAction(action)

        # putting the returned actions somewhere is necessary, otherwise
        # there will be none inside the single menus
        (self.channel_menu, self.variable_menu, self.function_menu), _ = (
            variables_handling.get_menus(self.insert_variable)
        )
        menu.addMenu(self.channel_menu)
        menu.addMenu(self.variable_menu)
        menu.addMenu(self.function_menu)
        # menu.addMenu(operator_menu)
        menu.addSeparator()
        (
            self.channel_menu2,
            self.variable_menu2,
            operator_menu2,
            self.function_menu2,
        ), __ = variables_handling.get_menus(self.append_variable, "Append")
        menu.addMenu(self.channel_menu2)
        menu.addMenu(self.variable_menu2)
        menu.addMenu(self.function_menu2)
        menu.addMenu(operator_menu2)

        # Connect the search bar to the filtering function
        search_bar.textChanged.connect(self.filter_actions)

        menu.exec_(self.mapToGlobal(pos))

    def filter_actions(self, query):
        for menu in [
            self.channel_menu,
            self.variable_menu,
            self.function_menu,
            self.channel_menu2,
            self.variable_menu2,
            self.function_menu2,
        ]:
            for action in menu.actions():
                action.setVisible(query.lower() in action.text().lower())

    def append_variable(self, val):
        """Used for the single actions of the context menu.

        Parameters
        ----------
        val :


        Returns
        -------

        """
        ind = self.table.selectedIndexes()[0]
        item = self.table_model.itemFromIndex(ind)
        text = item.text()
        item.setText(f"{text}{val}")

    def insert_variable(self, val):
        """Used for the single actions of the context menu.

        Parameters
        ----------
        val :


        Returns
        -------

        """
        ind = self.table.selectedIndexes()[0]
        item = self.table_model.itemFromIndex(ind)
        item.setText(f"{val}")

    def check_string(self, item):
        """If an element is part of the checkstrings, the item becomes
        green if valid, red otherwise and white if empty.

        Parameters
        ----------
        item :


        Returns
        -------

        """
        ind = item.index()
        pos = ind.column() if self.horizontal else ind.row()
        if pos not in self.checkstrings or item.text() == "":
            color = variables_handling.get_color("white")
        elif variables_handling.check_eval(item.text()):
            color = variables_handling.get_color("green")
        else:
            color = variables_handling.get_color("red")
        self.table_model.setData(
            self.table_model.indexFromItem(item), QBrush(color), Qt.BackgroundRole
        )

    def add(self, vals=None, change_focus=False):
        """Add the `vals` to the table as a new line. Checks are done
        for each part, whether there should be a comboBox etc.

        Parameters
        ----------
        vals :
            (Default value = None)

        Returns
        -------

        """
        if vals is None:
            vals = [""] * len(self.headerLabels) if len(self.headerLabels) else ""
            if self.default_values:
                for i, name in enumerate(self.headerLabels):
                    if name in self.default_values:
                        vals[i] = self.default_values[name]
        items = []
        box_indexes = []
        boxes = []
        table_indexes = []
        tables = []
        for i, name in enumerate(self.headerLabels):
            item = QStandardItem()
            if name in self.comboBoxes:
                box = QComboBox()
                for text in self.comboBoxes[name]:
                    box.addItem(text)
                self.boxes.append(box)
                box_indexes.append(i)
                boxes.append(box)
                if vals[i] in self.comboBoxes[name]:
                    box.setCurrentText(vals[i])
            elif name in self.subtables:
                if type(vals[i]) is not list:
                    vals[i] = []
                checksting = 0 if i in self.checkstrings else None
                table = AddRemoveTable(
                    horizontal=self.horizontal,
                    headerLabels=[],
                    tableData=vals[i],
                    growsize=False,
                    checkstrings=checksting,
                    fixedsize=True,
                )
                self.tables.append(table)
                table_indexes.append(i)
                tables.append(table)
                if vals[i] in self.subtables[name]:
                    table.setCurrentText(vals[i])
            elif i in self.checkables:
                item.setCheckable(True)
                item.setCheckState(
                    Qt.CheckState.Checked if vals[i] else Qt.CheckState.Unchecked
                )
            else:
                item = QStandardItem(str(vals[i]))
            item.setEditable(i in self.editables)
            item.setEnabled(i in self.enableds)
            items.append(item)
        if len(self.headerLabels) == 0:
            if self.comboBoxes:
                box = QComboBox()
                box.addItems(self.comboBoxes)
                self.boxes.append(box)
                box_indexes.append(0)
                boxes.append(box)
                if vals in self.comboBoxes:
                    box.setCurrentText(vals)
            item = QStandardItem(str(vals))
            items.append(item)
        if self.horizontal:
            self.table_model.appendRow(items)
            for j, i in enumerate(box_indexes):
                index = self.table_model.index(self.table_model.rowCount() - 1, i)
                self.table.setIndexWidget(index, boxes[j])
            for j, i in enumerate(table_indexes):
                index = self.table_model.index(self.table_model.rowCount() - 1, i)
                self.table.setIndexWidget(index, tables[j])
            # Set focus to the first editable cell in the newly added row
            focus_index = self.table_model.index(self.table_model.rowCount() - 1, 0)
            if change_focus:
                self.table.setCurrentIndex(focus_index)
                self.table.edit(focus_index)
        else:
            self.table_model.appendColumn(items)
            for j, i in enumerate(box_indexes):
                index = self.table_model.index(i, self.table_model.columnCount() - 1)
                self.table.setIndexWidget(index, boxes[j])
            for j, i in enumerate(table_indexes):
                index = self.table_model.index(i, self.table_model.columnCount() - 1)
                self.table.setIndexWidget(index, tables[j])
            # Set focus to the first editable cell in the newly added column
            focus_index = self.table_model.index(0, self.table_model.columnCount() - 1)
            if change_focus:
                self.table.setCurrentIndex(focus_index)
                self.table.edit(focus_index)
        for item in items:
            self.check_string(item)
        self.table.resizeColumnsToContents()
        self.table.resizeRowsToContents()
        self.update_max_hight()
        if self.horizontal:
            self.added.emit(items[0].row())
        else:
            self.added.emit(items[0].column())
        self.check_selection()


    def remove(self):
        """ """
        try:
            index = self.table.selectedIndexes()[0]
        except IndexError:
            raise Exception("You need to select a row first!")
        if self.askdelete:
            entry = self.table_model.itemFromIndex(index).text()
            remove_dialog = QMessageBox.question(
                self,
                "Remove entry?",
                f"Are you sure you want to remove the entry {entry}?",
                QMessageBox.Yes | QMessageBox.No,
            )
            if remove_dialog != QMessageBox.Yes:
                return
        row = index.row()
        col = index.column()
        if self.horizontal and row >= 0:
            self.table_model.removeRow(row)
            if not self.headerLabels and self.comboBoxes:
                self.boxes.pop(row)
            self.removed.emit(row)
        elif not self.horizontal and col >= 0:
            self.table_model.removeColumn(col)
            if not self.headerLabels and self.comboBoxes:
                self.boxes.pop(col)
            self.removed.emit(col)
        self.update_table_data()
        self.update_max_hight()
        self.check_selection()

    def update_table_data(self):
        """Reading all the data of the table, putting it as dict into
        `self.tableData` and returning it.

        Parameters
        ----------

        Returns
        -------

        """
        self.tableData = {}
        for i, lab in enumerate(self.headerLabels):
            vals = []
            self.tableData.update({lab: vals})
            if self.horizontal:
                for j in range(self.table_model.rowCount()):
                    ind = self.table_model.index(j, i)
                    if lab in self.comboBoxes:
                        vals.append(self.table.indexWidget(ind).currentText())
                    elif lab in self.subtables:
                        tab = self.table.indexWidget(ind)
                        tab.update_table_data()
                        vals.append(tab.tableData)
                    elif i in self.checkables:
                        vals.append(
                            self.table_model.item(j, i).checkState()
                            != Qt.CheckState.Unchecked
                        )
                    else:
                        try:
                            vals.append(int(self.table_model.item(j, i).text()))
                        except:
                            try:
                                vals.append(float(self.table_model.item(j, i).text()))
                            except:
                                vals.append(self.table_model.item(j, i).text())
            else:
                for j in range(self.table_model.columnCount()):
                    ind = self.table_model.index(i, j)
                    if lab in self.comboBoxes:
                        vals.append(self.table.indexWidget(ind).currentText())
                    elif lab in self.subtables:
                        tab = self.table.indexWidget(ind)
                        tab.update_table_data()
                        vals.append(tab.tableData)
                    else:
                        try:
                            vals.append(float(self.table_model.item(i, j).text()))
                        except:
                            vals.append(self.table_model.item(i, j).text())
        if len(self.headerLabels) == 0:
            self.tableData = []
            if self.comboBoxes:
                for box in self.boxes:
                    self.tableData.append(box.currentText())
            elif self.horizontal:
                for j in range(self.table_model.rowCount()):
                    try:
                        self.tableData.append(float(self.table_model.item(j, 0).text()))
                    except:
                        self.tableData.append(self.table_model.item(j, 0).text())
            else:
                for j in range(self.table_model.columnCount()):
                    try:
                        self.tableData.append(float(self.table_model.item(0, j).text()))
                    except:
                        self.tableData.append(self.table_model.item(0, j).text())
        return self.tableData


class AddRemoveDialoge(QDialog):
    """A QDialog providing an AddRemoveTable."""

    def __init__(
        self,
        addLabel="+",
        removeLabel="-",
        horizontal=True,
        editables=None,
        checkables=(),
        headerLabels=None,
        orderBy=None,
        parent=None,
        tableData=None,
        title="",
        comboBoxes=None,
        subtables=None,
        checkstrings=None,
        askdelete=False,
        default_values=None,
    ):
        super().__init__(parent)
        self.setWindowFlags(self.windowFlags() | Qt.WindowMaximizeButtonHint)
        self.buttonBox = QDialogButtonBox(self)
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")

        self.buttonBox.rejected.connect(self.reject)
        self.buttonBox.accepted.connect(self.accept)

        self.table = AddRemoveTable(
            addLabel=addLabel,
            removeLabel=removeLabel,
            horizontal=horizontal,
            editables=editables,
            checkables=checkables,
            headerLabels=headerLabels,
            orderBy=orderBy,
            parent=self,
            tableData=tableData,
            title=title,
            comboBoxes=comboBoxes,
            subtables=subtables,
            growsize=False,
            checkstrings=checkstrings,
            askdelete=askdelete,
            default_values=default_values,
        )

        layout = QGridLayout()
        self.setLayout(layout)
        layout.addWidget(self.table, 0, 0)
        layout.addWidget(self.buttonBox, 1, 0)
        self.setWindowTitle(title)
        self.table.sizechange.connect(self.adjustSize)
        self.adjustSize()
        self.setMinimumWidth(len(headerLabels) * 70)

    def get_data(self):
        """ """
        return self.table.update_table_data()

    def keyPressEvent(self, a0: QKeyEvent) -> None:
        """Overwrites the keyPressEvent of the QDialog so that it does
        not close when pressing Enter/Return.

        Parameters
        ----------
        a0: QKeyEvent :


        Returns
        -------

        """
        if a0.key() == Qt.Key_Enter or a0.key() == Qt.Key_Return:
            return
        super().keyPressEvent(a0)


class MultiTableDialog(QDialog):
    """A QDialog providing an AddRemoveTable."""

    def __init__(
        self,
        table_args=[],  # list of dicts with arguments for AddRemoveTable
        parent=None,
        title="",
    ):
        super().__init__(parent)
        self.setWindowFlags(self.windowFlags() | Qt.WindowMaximizeButtonHint)
        self.buttonBox = QDialogButtonBox(self)
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")

        self.buttonBox.rejected.connect(self.reject)
        self.buttonBox.accepted.connect(self.accept)

        self.tables = []
        max_labels = 0
        for args in table_args:
            table = AddRemoveTable(**args)
            table.sizechange.connect(self.adjustSize)
            self.tables.append(table)
            if "headerLabels" in args:
                max_labels = max(max_labels, len(args["headerLabels"]))

        layout = QGridLayout()
        self.setLayout(layout)
        for i, table in enumerate(self.tables):
            layout.addWidget(table, i, 0)
        layout.addWidget(self.buttonBox, len(self.tables), 0)
        self.setWindowTitle(title)
        self.adjustSize()
        self.setMinimumWidth(max_labels * 100)

    def get_data(self):
        """ """
        data = []
        for table in self.tables:
            data.append(table.update_table_data())
        return data

    def keyPressEvent(self, a0: QKeyEvent) -> None:
        """Overwrites the keyPressEvent of the QDialog so that it does
        not close when pressing Enter/Return.

        Parameters
        ----------
        a0: QKeyEvent :


        Returns
        -------

        """
        if a0.key() == Qt.Key_Enter or a0.key() == Qt.Key_Return:
            return
        super().keyPressEvent(a0)

    def hide_add_remove_buttons(self):
        """ """
        for table in self.tables:
            table.addButton.hide()
            table.removeButton.hide()



"""Functions to be used in protocols.

The functions in this module may be used inside a protocol. It is much
simpler to use these than writing them as a string into the protocol-file.
"""

import numpy as np
from bluesky import plan_stubs as bps
from bluesky.utils import Msg

from ophyd import SignalRO, Device

from PySide6.QtWidgets import (
    QMessageBox,
    QWidget,
    QDialog,
    QDialogButtonBox,
    QGridLayout,
    QLabel,
    QLineEdit,
    QProgressBar,
    QPushButton,
    QComboBox,
    QTextEdit,
    QCheckBox,
)
from PySide6.QtCore import Signal, Qt
from PySide6.QtGui import QFont, QIcon

from nomad_camels.ui_widgets.add_remove_table import AddRemoveTable
from nomad_camels.ui_widgets.channels_check_table import Channels_Check_Table
from nomad_camels import graphics

from suitcase.nomad_camels_hdf5 import Serializer, export

import inspect
import os
import sys
import glob

from importlib import resources

from nomad_camels.bluesky_handling.loop_step_functions_api_call import (
    execute_camels_api_call,
    execute_generic_api_call,
    save_API_response_to_variable,
    evaluate_message_body,
)


def get_newest_file(directory):
    """
    Return the newest file in the given directory.

    Parameters
    ----------
    directory : str
        The directory where the newest file should be found.

    Returns
    -------
    newest_file : str
        The path to the newest file in the directory.
    """
    # If the given directory is a file, take the directory of the file
    if os.path.isfile(directory):
        directory = os.path.dirname(directory)
    # List all files in the directory
    files = glob.glob(os.path.join(directory, "*"))
    # Find the newest file
    newest_file = max(files, key=os.path.getmtime)
    return newest_file


def export_function(
    runs,
    save_path,
    do_export,
    new_file_each=True,
    export_csv=False,
    export_json=False,
    plot_data=None,
    do_nexus_output=True,
):
    """
    Export the given runs to the given path.

    Parameters
    ----------
    runs : list[databroker.core.BlueskyRun]
        The runs to be exported.
    save_path : str
        The path where the files should be saved.
    do_export : bool
        Whether the runs should be exported.
    new_file_each : bool, optional
        (Default value = True)
        Whether a new file should be created for each run. If True, the files are
        padded with the start-time of the run. Ignored if `do_export` is False.
    export_csv : bool, optional
        (Default value = False)
        Whether the data should be exported to a csv file.
    export_json : bool, optional
        (Default value = False)
        Whether the metadata should be exported to a json file.
    plot_data : dict, optional
        (Default value = None)
        The data to be plotted.

    """
    if os.path.splitext(save_path)[1] == ".nxs":
        fname = os.path.basename(save_path)
        path = os.path.dirname(save_path)
    if do_export:
        if not isinstance(runs, list):
            runs = [runs]
        for run in runs:
            docs = run.documents(fill="yes")
            export(
                docs,
                path,
                fname,
                new_file_each,
                plot_data=plot_data,
                do_nexus_output=do_nexus_output,
            )
    if export_csv or export_json:
        from nomad_camels.utility.databroker_export import export_h5_to_csv_json

        file = get_newest_file(path)
        export_h5_to_csv_json(file, export_data=export_csv, export_metadata=export_json)


def saving_function(
    name, start_doc, path, new_file_each=True, plot_data=None, do_nexus_output=True
):
    """
    Create a Serializer and return it.

    The serializer is from the suitcase.nomad_camels_hdf5 module and used to save
    the data during a run.

    Parameters
    ----------
    name : str
        The name of the file.
    start_doc : dict
        The start document of the run.
    path : str
        The path where the file should be saved.
    new_file_each : bool, optional
        (Default value = True)
        Whether a new file should be created for each run. If True, the files are
        padded with the start-time of the run.
    plot_data : dict, optional
        (Default value = None)
        The data to be plotted.

    Returns
    -------
    serializers : list
        The serializers to be used.
    resources : list
        The resources to be used. In this case, an empty list.
    """
    fname = start_doc["uid"]
    if os.path.splitext(path)[1] == ".nxs":
        fname = os.path.basename(path)
        path = os.path.dirname(path)
    return [
        Serializer(
            path,
            fname,
            new_file_each=new_file_each,
            plot_data=plot_data,
            do_nexus_output=do_nexus_output,
        )
    ], []


def trigger_multi(devices, grp=None):
    """
    Trigger multiple devices.

    Parameters
    ----------
    devices : list[ophyd.Device]
        List of the devices that should be triggered.

    grp : string (or any hashable object), optional
        identifier used by 'wait'; None by default
    """
    for obj in devices:
        if hasattr(obj, "trigger"):
            yield from bps.trigger(obj, group=grp)


def read_wo_trigger(devices, grp=None, stream="primary"):
    """
    Used if not reading by trigger_and_read, but splitting both. This function only reads, without triggering.

    Parameters
    ----------
    devices : list[ophyd.Device]
        List of the devices that should be read.
    grp : string (or any hashable object), optional
        identifier used by 'wait'; None by default
    stream : string, optional
        event stream name, a convenient human-friendly identifier; default
        name is 'primary'

    Returns
    -------
    ret : dict
        The readings of the devices.
    """
    if grp is not None:
        yield from bps.wait(grp)
    yield from bps.create(stream)
    ret = {}  # collect and return readings to give plan access to them
    for obj in devices:
        reading = yield from bps.read(obj)
        if reading is not None:
            ret.update(reading)
    yield from bps.save()
    return ret


def simplify_configs_dict(configs):
    """
    Returns a simplified version of the given dictionary `configs` by returning confs[key] = configs[key]['value'].

    Parameters
    ----------
    configs : dict
        The dictionary to be simplified.


    Returns
    -------
    confs : dict
        The simplified dictionary.
    """
    confs = {}
    for key, value in configs.items():
        if "value" in value:
            val = value["value"]
        else:
            val = value
        if isinstance(val, Device):
            confs[key] = val.name
        else:
            confs[key] = val
    return confs


def get_fit_results(fits, namespace, yielding=False, stream="primary"):
    """
    Updates and reads all the fits that correspond to the given stream and
    resets the fits in the end.

    Parameters
    ----------
    fits : dict
        Dictionary of all the fits to take into account.
    namespace : dict
        Namespace dictionary where to write the fit results.
    yielding : bool, optional
         (Default value = False)
         If True, the fits will be triggered and updated.
    stream : str, optional
         (Default value = 'primary')
         The stream on which the regarded fits should run. Only the fits which
         have `stream_name` equal to `stream` will be used.
    """
    for name, fit in fits.items():
        if yielding and fit.stream_name == stream:
            yield from bps.trigger_and_read(
                [fit.read_ready], name=f"{stream}_fits_readying_{name}"
            )
            yield from fit.update_fit()
            # yield from bps.trigger_and_read(fit.ophyd_fit.used_comps,
            #                                 name=f'{stream}_fits_{name}')
        if not fit.result:
            continue
        for param in fit.params:
            if param in fit.result.best_values:
                namespace[f"{name}_{param}"] = fit.result.best_values[param]
        namespace[f"{name}_covar"] = fit.result.covar
        fit._reset()


def make_recoursive_plot_list_of_sub_steps(sub_dict, plot_list=None):
    plot_list = plot_list or []
    for key, value in sub_dict.items():
        if key == "plots":
            plot_list += value
        elif isinstance(value, dict):
            plot_list = make_recoursive_plot_list_of_sub_steps(value, plot_list)
    return plot_list


def clear_plots(plots, stream="primary"):
    """
    Clears all given plots if they correspond to the given stream.

    Parameters
    ----------
    plots : list
        List of the plots to be cleared.
    stream : str
         (Default value = 'primary')
         The stream to which the plots that should be cleared correspond.
    """
    for plot in plots:
        if plot.stream_name == stream or plot.stream_name.startswith(f"{stream}_fits_"):
            plot.clear_plot()


def gradient_descent(
    max_iterations,
    threshold,
    w_init,
    func_text,
    evaluator,
    set_channel,
    read_channels,
    min_step,
    max_step,
    min_val,
    max_val,
    stream_name="gradient_descent",
    learning_rate=0.05,
    momentum=0.8,
    max_step_for_diff=None,
):
    """
    Helper function for the gradient descent protocol-step.
    It follows a simple gradient descent algorithm to find an optimum.

    Parameters
    ----------
    max_iterations : int
        The maximum number of iterations until the algorithm should stop if it
        did not arrive at the threshold yet.
    threshold : float
        If the difference between two measurements is smaller than this
        threshold, the algorithm recognizes the value as the optimum and stops.
    w_init : float
        The initial set-value from where the algorithm should start.
    func_text : str
        This string is evaluated by the given evaluator to give the target
        function.
    evaluator : Evaluator
        Used to evaluate the read values.
    set_channel : ophyd.Signal
        The channel which is used for the optimization.
    read_channels : list
        A list of all the channels which are read for the optimization.
    min_step : float
        The minimum step size.
    max_step : float
        The maximum step size.
    min_val : float
        The minimum value that should be given to the `set_channel`.
    max_val : float
        The maximum value that should be given to the `set_channel`.
    stream_name : str
         (Default value = 'gradient_descent')
         The bluesky stream in which everything should run.
    learning_rate : float
         (Default value = 0.05)
         A weight for the learning of the gradient descent.
         The next shift `delta_w` is calculated as:
         delta_w = -learning_rate * <current_gradient> + momentum * <last_delta_w>
    momentum : float
         (Default value = 0.8)
         A momentum to keep up the last direction.
         The next shift `delta_w` is calculated as:
         delta_w = -learning_rate * <current_gradient> + momentum * <last_delta_w>
    max_step_for_diff : float, None
        If none, max_step_for_diff = 10 * min_step
        Only if the last step was smaller than `max_step_for_diff`, the
        algorithm breaks if the threshold is reached.

    Returns
    -------
    w_history :
        The history of values of set_channel
    f_history :
        The history of evaluated values

    """
    if set_channel not in read_channels:
        read_channels += [set_channel]

    def obj_func(set_val):
        """
        Set the value and then read the channels for each iteration of the algorithm.

        Parameters
        ----------
        set_val : float
            Value to be set on the set_channel
        """
        yield from bps.checkpoint()
        yield from bps.abs_set(set_channel, set_val, group="A")
        yield from bps.wait("A")
        yield from bps.trigger_and_read(read_channels, name=stream_name)
        # yield from bps.sleep(1)

    if max_step_for_diff is None:
        max_step_for_diff = 10 * min_step

    w = w_init
    w_history = [w]
    yield from obj_func(w)
    f_history = [evaluator.eval(func_text)]
    if w - min_val > max_val - w:
        w -= max_step_for_diff
        delta_w = -max_step_for_diff
    else:
        w += max_step_for_diff
        delta_w = max_step_for_diff
    i = 0
    diff = np.inf

    while i < max_iterations and (
        diff > threshold or np.abs(delta_w) > max_step_for_diff
    ):
        yield from obj_func(w)
        f = evaluator.eval(func_text)
        grad = (f - f_history[-1]) / (w - w_history[-1])
        delta_w = -learning_rate * grad + momentum * delta_w
        w_history.append(w)
        f_history.append(f)
        if np.abs(delta_w) > max_step:
            delta_w = np.sign(delta_w) * max_step
        elif np.abs(delta_w) < min_step:
            delta_w = np.sign(delta_w) * min_step
        w += delta_w
        if w < min_val:
            w = min_val
        elif w > max_val:
            w = max_val
        if w == w_history[-1]:
            if w - min_val > max_val - w:
                w -= max_step_for_diff
                delta_w -= max_step_for_diff
            else:
                w += max_step_for_diff
                delta_w = max_step_for_diff
        # store the history of w and f

        # update iteration number and diff between successive values
        # of objective function
        i += 1
        diff = np.abs(f_history[-1] - f_history[-2])

    sort_w = [i for j, i in sorted(zip(f_history, w_history))]
    sort_f = sorted(f_history)
    while True:
        fs = []
        for i in range(2):
            yield from obj_func(sort_w[0])
            fs.append(evaluator.eval(func_text))
        if fs[0] and fs[1] < sort_f[0] + 0.1 * (sort_f[-1] - sort_f[0]):
            break
        elif len(sort_f) < 0.5 * len(f_history):
            yield from obj_func(w_init)
            break
        else:
            sort_f.pop(0)
            sort_w.pop(0)
    return w_history, f_history


def get_range(
    evaluator,
    start,
    stop,
    points: int,
    min_val=np.nan,
    max_val=np.nan,
    loop_type="start - stop",
    sweep_mode="linear",
    endpoint=True,
    distance=np.nan,
    use_distance=False,
):
    """
    Helper function for steps like sweeps and the for loop.

    Using the evaluator, it creates the range of iterations given the other
    values.

    Parameters
    ----------
    evaluator : Evaluator
        Used to evaluate the given expressions (`start`, `stop`, `points`,
        `min_val`, `max_val`).
    start : float, str
        The starting position for the sweep / loop.
    stop : float, str
        The end position for the sweep / loop.
    points : int, str
        The number of points in the range. If using `min_val` / `max_val`,
        `points` is the distance between those two.
    min_val : float, str
         (Default value = np.nan)
         Minimum value of the loop / sweep. Useful for hysteresis-sweeps.
    max_val : float, str
         (Default value = np.nan)
         Minimum value of the loop / sweep. Useful for hysteresis-sweeps.
    loop_type : str
         (Default value = 'start - stop')
         Possible values: 'start - stop', ignoring `min_val` and `max_val`
         'start - min - max - stop' going to minimum first, then maximum
         'start - max - min - stop', or anything not-specified.
    sweep_mode : str
         (Default value = 'linear')
         The type how the distance between the points is calculated.
         If 'linear', they are equidistant, if 'logarithmic', the points between
         log(start) and log(stop) are equidistant, if 'exponential', between
         exp(start) and exp(stop), otherwise they are equidistant between
         1/start and 1/stop.
    endpoint : bool
         (Default value = True)
         Whether to include the endpoint into the range.
    distance : float, str
        (Default value = np.nan)
        The distance between the points.
    use_distance : bool
        (Default value = False)
        If True, the distance is used to calculate the range.

    Returns
    -------
        An array of the calculated range.
    """
    start = float(evaluator.eval(str(start)))
    stop = float(evaluator.eval(str(stop)))
    points = int(evaluator.eval(str(points)))
    min_val = evaluator.eval(str(min_val))
    max_val = evaluator.eval(str(max_val))
    try:
        distance = evaluator.eval(str(distance))
    except ValueError:
        distance = np.nan
    if loop_type == "start - stop":
        return get_inner_range(
            start,
            stop,
            points,
            sweep_mode,
            endpoint,
            use_distance=use_distance,
            distance=distance,
        )
    if (
        min_val > max_val
        or not (min_val <= start <= max_val)
        or not (min_val <= stop <= max_val)
    ):
        raise ValueError("Start and Stop must be between min and max!")
    full_dist = np.abs(max_val - min_val)
    if loop_type == "start - min - max - stop":
        full_dist += np.abs(start - min_val) + np.abs(stop - max_val)
        p1 = round(points * np.abs(start - min_val) / full_dist)
        p3 = round(points * np.abs(stop - max_val) / full_dist)
        p2 = points - p1 - p3
        vals1 = get_inner_range(
            start, min_val, p1, sweep_mode, endpoint, use_distance, distance
        )
        vals2 = get_inner_range(
            min_val, max_val, p2, sweep_mode, endpoint, use_distance, distance
        )
        vals3 = get_inner_range(
            max_val, stop, p3, sweep_mode, endpoint, use_distance, distance
        )
    else:
        full_dist += np.abs(start - max_val) + np.abs(stop - min_val)
        p1 = round(points * np.abs(start - max_val) / full_dist)
        p3 = round(points * np.abs(stop - min_val) / full_dist)
        p2 = points - p1 - p3
        vals1 = get_inner_range(
            start, max_val, p1, sweep_mode, endpoint, use_distance, distance
        )
        vals2 = get_inner_range(
            max_val, min_val, p2, sweep_mode, endpoint, use_distance, distance
        )
        vals3 = get_inner_range(
            min_val, stop, p3, sweep_mode, endpoint, use_distance, distance
        )
    return np.concatenate([vals1, vals2, vals3])


def get_inner_range(
    start, stop, points, sweep_mode, endpoint, use_distance=False, distance=np.nan
):
    """
    Used for `get_range`, to make the split up ranges if doing a hysteresis
    sweep.

    Parameters
    ----------
    start : float, str
        The starting position for the sweep / loop.
    stop : float, str
        The end position for the sweep / loop.
    points : float, str
        The number of points in the range. If using `min_val` / `max_val`,
        `points` is the distance between those two.
    sweep_mode : str
         (Default value = 'linear')
         The type how the distance between the points is calculated.
         If 'linear', they are equidistant, if 'logarithmic', the points between
         log(start) and log(stop) are equidistant, if 'exponential', between
         exp(start) and exp(stop), otherwise they are equidistant between
         1/start and 1/stop.
    endpoint : bool
         (Default value = True)
         Whether to include the endpoint into the range.

    Returns
    -------
        An array of the calculated range.
    """
    if sweep_mode == "linear":
        if use_distance:
            if start < stop:
                vals = np.arange(start, stop, distance)
            else:
                vals = np.arange(start, stop, -distance)
            if endpoint:
                vals = np.append(vals, stop)
            return vals
        return np.linspace(start, stop, points, endpoint=endpoint)
    elif sweep_mode == "logarithmic":
        start = np.log(start)
        stop = np.log(stop)
        return np.exp(np.linspace(start, stop, points, endpoint=endpoint))
    elif sweep_mode == "exponential":
        start = np.exp(start)
        stop = np.exp(stop)
        return np.log(np.linspace(start, stop, points, endpoint=endpoint))
    return 1 / np.linspace(1 / start, 1 / stop, points, endpoint=endpoint)


class Prompt_Box(QMessageBox):
    """
    A subclass of QMessageBox that is used in the prompt-step.
    The protocol is paused until `self.done_flag` is True.

    Parameters
    ----------
    icon : str
        If 'Error', the `QMessagebox.Critical` icon is displayed, if 'Warning',
        then `QMessagebox.Warning` is used, otherwise `QMessagebox.Information`.
    text : str
        The text to be displayed by the prompt.
    title : str
        The window-title of the prompt.
    """

    def __init__(self, icon="", text="", title="", parent=None, abortable=False):
        super().__init__(parent=parent)
        if icon == "Error":
            self.setIcon(QMessageBox.Critical)
        elif icon == "Warning":
            self.setIcon(QMessageBox.Warning)
        else:
            self.setIcon(QMessageBox.Information)
        self.setText(text)
        self.setWindowTitle(title)
        self.helper = BoxHelper()
        self.helper.executor.connect(self.start_execution)
        self.buttonClicked.connect(self.set_done)
        self.abort_flag = False
        if abortable:
            self.ok_button = QPushButton("OK")
            self.addButton(self.ok_button, QMessageBox.ButtonRole.AcceptRole)
            self.ok_button.clicked.connect(self.set_done)
            self.abort_button = QPushButton("Abort Protocol")
            self.addButton(self.abort_button, QMessageBox.ButtonRole.NoRole)
            self.abort_button.clicked.connect(self.abort_action)
        self.done_flag = False

    def set_done(self):
        """Sets `self.done_flag` to True."""
        self.done_flag = True

    def start_execution(self):
        """Sets `self.done_flag` to False and starts `self.exec()`."""
        self.done_flag = False
        self.exec()

    def abort_action(self):
        """
        If the abort button is clicked, the protocol is stopped.
        """
        self.abort_flag = True


class BoxHelper(QWidget):
    """Helper-class to start the execution of Prompts and other boxes from
    within the protocol."""

    executor = Signal()


class Value_Box(QDialog):
    """
    Dialog to set variables or channels at runtime of a protocol.

    Parameters
    ----------
    text : str
        A text to be displayed with the box.
    title : str
        The window title of the box.
    variables : list, default: None
        The variables that should be set by the user. A QLineEdit will be
        provided for each of the variables.
    channels : list, default: None
        The channels that should be set by the user. A QLineEdit will be
        provided for each of the channels.
    free_variables : bool, default: False
        Whether the user is allowed to freely set any variables.
    free_channels : bool, default: False
        Whether the user is allowed to freely set any channels.
    devs : dict, default: None
        Dictionary of the available devices. Only needed, if `free_channels` is
        True, to provide the available channels.
    """

    def __init__(
        self,
        text="",
        title="",
        variables=None,
        channels=None,
        free_variables=False,
        free_channels=False,
        parent=None,
        devs=None,
        comboboxes=None,
    ):
        super().__init__(parent)
        self.setWindowFlags(self.windowFlags() | Qt.WindowMaximizeButtonHint)
        text_label = QLabel(text)
        self.buttonBox = QDialogButtonBox(self)
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.buttonBox.setToolTip(
            '"OK" will set the values, click "Cancel" to set no new values.'
        )
        layout = QGridLayout()

        self.helper = BoxHelper()
        self.helper.executor.connect(self.start_execution)
        self.done_flag = False

        channels = channels or []
        n_channels = len(channels)
        self.channel_boxes = []
        self.channels = channels
        for i, channel in enumerate(channels):
            channel_label = QLabel(f"{channel}:")
            channel_box = QLineEdit()
            self.channel_boxes.append(channel_box)
            layout.addWidget(channel_label, 2 + i, 0)
            layout.addWidget(channel_box, 2 + i, 1)

        variables = variables or []
        n_variables = len(variables)
        self.variable_boxes = []
        self.variables = variables
        for i, variable in enumerate(variables):
            variable_label = QLabel(f"{variable}:")
            if comboboxes and comboboxes[i]:
                variable_box = QComboBox()
                variable_box.addItems([str(x) for x in comboboxes[i]])
            else:
                variable_box = QLineEdit()
            self.variable_boxes.append(variable_box)
            layout.addWidget(variable_label, 2 + i, 2)
            layout.addWidget(variable_box, 2 + i, 3)

        font = QFont()
        font.setBold(True)
        if channels:
            channels_label = QLabel("Channels")
            channels_label.setStyleSheet("font-size: 9pt")
            channels_label.setFont(font)
            layout.addWidget(channels_label, 1, 0, 1, 2)
        if variables:
            variables_label = QLabel("Variables")
            variables_label.setStyleSheet("font-size: 9pt")
            variables_label.setFont(font)
            layout.addWidget(variables_label, 1, 2, 1, 2)

        devs = devs or {}
        self.channel_devs = {}
        for dev, val in devs.items():
            self.channel_devs.update(get_channels(val))
        self.channel_table = None
        self.variable_table = None
        if free_channels:
            self.channel_table = Channels_Check_Table(
                self,
                ["set", "channel", "value"],
                True,
                channels=list(self.channel_devs.keys()),
            )
            layout.addWidget(self.channel_table, 2 + n_channels + n_variables, 0, 1, 2)
        if free_variables:
            header = ["variable", "value"]
            self.variable_table = AddRemoveTable(headerLabels=header)
            layout.addWidget(self.variable_table, 2 + n_channels + n_variables, 2, 1, 2)

        self.buttonBox.rejected.connect(self.reject)
        self.buttonBox.accepted.connect(self.accept)

        self.setLayout(layout)
        layout.addWidget(text_label, 0, 0, 1, 4)
        layout.addWidget(self.buttonBox, 10 + n_channels + n_variables, 0, 1, 4)
        self.setWindowTitle(title)
        # self.table.sizechange.connect(self.adjustSize)
        self.adjustSize()
        self.was_accepted = False
        self.set_variables = {}
        self.set_channels = {}

    def start_execution(self):
        """Sets `self.done_flag` to False and starts `self.exec()`."""
        self.done_flag = False
        self.exec()

    def accept(self) -> None:
        """
        Reads all the values to be set to channels / variables and saves them in
        `self.set_channels` and `self.set_variables` respectively, then sets
        `self.done_flag` to True, allowing the protocol to go on and accepts the dialog.
        """
        self.set_variables = {}
        self.set_channels = {}
        for i, v_box in enumerate(self.variable_boxes):
            if isinstance(v_box, QLineEdit):
                val = v_box.text()
            else:
                val = v_box.currentText()
            if val:
                self.set_variables[self.variables[i]] = val
        for i, c_box in enumerate(self.channel_boxes):
            val = c_box.text()
            if val:
                self.set_channels[self.channels[i]] = val
        if self.variable_table:
            var_table_data = self.variable_table.update_table_data()
            for i, var in enumerate(var_table_data["variable"]):
                if var in self.set_variables:
                    box = Prompt_Box(
                        "Critical",
                        f"Cannot set two values to variable {var}!",
                        "identical variable names",
                    )
                    box.exec()
                    return
                self.set_variables[var] = var_table_data["value"][i]
        if self.channel_table:
            channel_table_data = self.channel_table.get_info()
            for i, channel in enumerate(channel_table_data["channel"]):
                if channel in self.set_channels:
                    box = Prompt_Box(
                        "Critical",
                        f"Cannot set two values to channel {channel}!",
                        "identical channel names",
                    )
                    box.exec()
                    return
                self.set_channels[channel] = channel_table_data["value"][i]
        self.was_accepted = True
        self.done_flag = True
        return super().accept()

    def reject(self) -> None:
        """
        Sets `self.done_flag` to True, allowing the protocol to go on before
        rejecting the dialog.
        """
        self.done_flag = True
        return super().reject()


def get_channels(dev):
    """
    Goes through the components of the given ophyd device and returns all that
    are not read-only and not config.

    Parameters
    ----------
    dev : ophyd.Device
        The device that is checked for output channels.

    Returns
    -------
    channels : dict
        The keys are the channels in CAMELS-style. The values are
        [dev.name, <name_of_the_component>].
    """
    channels = {}
    for comp in dev.walk_components():
        if issubclass(comp.item.cls, SignalRO):
            continue
        name = comp.item.attr
        if name not in dev.configuration_attrs:
            channels[f"{dev.name}_{name}"] = [dev.name, name]
    return channels


def get_opyd_and_py_file_contents(classname, md, device_name):
    """
    Reads the content of the .py and _opyd.py driver files of the given instrument and adds
    them to the metadata dictionary. If the files are not found, it adds a corresponding message
    to the dictionary.

    Parameters
    ----------
    classname : class or function
        The class or function whose associated files' contents are to be read.
    md : dict
        The metadata dictionary where the content of the files should be added.
    device_name : str
        The name of the device associated with the classname.

    Returns
    -------
    md : dict
        The updated metadata dictionary with the content of the files added or with a message
        indicating that the file was not found.

    Raises
    ------
    FileNotFoundError
        If the .py or _opyd.py files associated with the classname are not found.
    """
    if inspect.isfunction(classname):
        try:
            with open(classname.__code__.co_filename, "r") as file:
                md["devices"][device_name][
                    f"python_file_{os.path.splitext(os.path.basename(classname.__code__.co_filename))[0]}"
                ] = file.read()
        except FileNotFoundError:
            md["devices"][device_name][
                os.path.splitext(os.path.basename(classname.__code__.co_filename))[0]
            ] = "._ophyd file not found"
        try:
            with open(
                classname.__code__.co_filename.replace("_ophyd.py", ".py"), "r"
            ) as file:
                md["devices"][device_name][
                    f"python_file_{os.path.splitext(os.path.basename(classname.__code__.co_filename))[0]}".replace(
                        "_ophyd", ""
                    )
                ] = file.read()
        except FileNotFoundError:
            md["devices"][device_name][
                os.path.splitext(os.path.basename(classname.__code__.co_filename))[
                    0
                ].replace("_ophyd", "")
            ] = ".py file not found"

    elif inspect.isclass(classname):
        try:
            with open(sys.modules[classname.__module__].__file__, "r") as file:
                md["devices"][device_name][
                    f"python_file_{os.path.splitext(os.path.basename(sys.modules[classname.__module__].__file__))[0]}"
                ] = file.read()
        except FileNotFoundError:
            md["devices"][device_name][
                os.path.splitext(
                    os.path.basename(sys.modules[classname.__module__].__file__)
                )[0]
            ] = "._ophyd file not found"
        try:
            with open(
                sys.modules[classname.__module__].__file__.replace("_ophyd.py", ".py"),
                "r",
            ) as file:
                md["devices"][device_name][
                    f"python_file_{os.path.splitext(os.path.basename(sys.modules[classname.__module__].__file__))[0]}".replace(
                        "_ophyd", ""
                    )
                ] = file.read()
        except FileNotFoundError:
            md["devices"][device_name][
                os.path.splitext(
                    os.path.basename(sys.modules[classname.__module__].__file__)
                )[0].replace("_ophyd", "")
            ] = ".py file not found"
    return md


def create_venv_run_file_delete_venv(packages, script_to_run):
    """
    Create a virtual environment, install the specified packages
    in it, run the specified Python script, and then delete the virtual environment.
    """
    import venv
    import subprocess
    import tempfile
    import shutil

    # Step 1: Create the virtual environment in a temp folder using tempfile
    try:
        temp_folder = tempfile.mkdtemp()
    except Exception as e:
        raise Exception(f"Error creating temporary folder: {e}") from e
    try:
        venv.create(temp_folder, with_pip=True)
    except Exception as e:
        raise Exception(f"Error creating virtual environment: {e}") from e

    # Path to the virtual environment's Python interpreter
    python_executable = (
        os.path.join(temp_folder, "bin", "python")
        if os.name != "nt"
        else os.path.join(temp_folder, "Scripts", "python.exe")
    )

    # Step 2: Install the specified packages
    for package, version in zip(packages["Python Package"], packages["Version"]):
        if not version:
            try:
                subprocess.run(
                    [python_executable, "-m", "pip", "install", f"{package}"]
                )
            except Exception as e:
                raise Exception(f"Error installing package {package}: {e}") from e
        else:
            try:
                subprocess.run(
                    [python_executable, "-m", "pip", "install", f"{package}=={version}"]
                )
            except Exception as e:
                raise Exception(
                    f"Error installing package {package}=={version}: {e}"
                ) from e

    # Step 3: Run the specified Python script
    try:
        result_python_file = subprocess.run(
            [python_executable, script_to_run], cwd=os.path.dirname(script_to_run)
        )
    except Exception as e:
        raise Exception(f"Error running script {script_to_run}: {e}") from e

    # Step 4: Delete the virtual environment
    try:
        if os.path.exists(temp_folder):
            shutil.rmtree(temp_folder)
    except Exception as e:
        raise Exception(f"Error deleting temporary folder: {e}") from e
    return result_python_file


def evaluate_python_file_output(stdout, namespace):
    """Evaluates the stdout of a Python file execution and writes the returned key value pairs to the namespace, so CAMELS can access their values"""
    import json
    import re

    if stdout:
        json_pattern = re.compile(r"###Start Data(.*?)###End Data", re.DOTALL)

        # Search for JSON in the stdout
        match = json_pattern.search(stdout)

        if match:
            json_str = match.group(1).strip()

            # Load the JSON string into a Python dictionary
            data = json.loads(json_str)
        else:
            raise ValueError("No valid dictionary found in the output")
        variables_dict = {}
        if isinstance(data, list):  # Step 1: Check if the stdout is a dictionary
            for item in data:  # Step 2: Loop through the list
                key, value = map(
                    str.strip, item.split("=", 1)
                )  # Step 3: Split each string
                variables_dict[key] = value

            for key, value in variables_dict.items():
                namespace[key] = value
                print(f"Python File returned:\n{key} = {value}")

        if isinstance(data, dict):
            for key, value in data.items():
                namespace[key] = value
                print(f"Python File returned:\n{key} = {value}")


class Value_Setter(QWidget):
    """A widget to set a value and a wait-time for a waiting bar."""

    set_signal = Signal(float)
    hide_signal = Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.start_time = 0
        self.end_time = 0
        self.timer = 0
        self.wait_time = 0

    def set_wait_time(self, wait_time):
        """Set the wait time for the waiting bar."""
        import datetime as dt

        self.wait_time = dt.timedelta(seconds=wait_time).total_seconds()

    def update_timer(self):
        """Update the timer of the waiting bar."""
        import datetime as dt

        self.timer = (dt.datetime.now() - self.start_time).total_seconds()
        self.set_signal.emit(self.timer / self.wait_time * 100)


class Waiting_Bar(QWidget):
    """A widget to show a waiting bar.

    Used in the wait step with a progress bar or also to wait for a condition and
    display the list plot of the condition.

    Parameters
    ----------
    parent : QWidget
        The parent widget.
    title : str
        The title of the waiting bar.
    skipable : bool
        (Default value = True)
        Whether the waiting bar can be skipped and shows a skip button.
    with_timer : bool
        (Default value = False)
        Whether the waiting bar has a timer.
    display_bar : bool
        (Default value = True)
        Whether the waiting bar should be displayed.
    plot : QWidget, optional
        A plot to be displayed together with the waiting bar.
    """

    def __init__(
        self,
        parent=None,
        title="",
        skipable=True,
        with_timer=False,
        display_bar=True,
        plot=None,
    ):
        super().__init__(parent=parent)
        layout = QGridLayout()
        self.progressBar = QProgressBar()
        self.progressBar.setValue(0)
        layout.addWidget(self.progressBar, 0, 0)
        self.progressBar.setHidden(not display_bar)

        self.skip = False
        if skipable:
            self.skipButton = QPushButton("SKIP")
            layout.addWidget(self.skipButton, 0, 1)
            self.skipButton.clicked.connect(self.skipping)
        self.setLayout(layout)
        self.setWindowTitle(title or "CAMELS progress bar")
        self.adjustSize()
        self.helper = BoxHelper()
        self.helper.executor.connect(self.start_execution)
        self.setter = Value_Setter()
        self.setter.set_signal.connect(self.setValue)
        self.setter.hide_signal.connect(self.hide)
        self.with_timer = with_timer
        if plot is not None:
            self.plot = plot
            layout.addWidget(self.plot, 1, 0, 1, 2)
        self.adjustSize()
        self.setWindowFlags(self.windowFlags() & ~Qt.WindowCloseButtonHint)

    def setValue(self, value):
        """Set the value of the progress bar."""
        self.progressBar.setValue(value)

    def skipping(self):
        """Set the skip flag to True and hide the waiting bar."""
        self.skip = True
        self.hide()

    def start_execution(self):
        """Sets `self.done_flag` to False and starts `self.exec()`."""
        self.skip = False
        self.setHidden(False)
        if self.with_timer:
            import datetime as dt

            self.setter.start_time = dt.datetime.now()

        self.show()


class Commenting_Box(QWidget):
    """A widget to add comments to the protocol."""

    closing = Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        layout = QGridLayout()
        self.comment_box = QTextEdit()
        self.finished_box = QCheckBox("commenting finished")
        self.setWindowTitle("Live Measurement Comments - NOMAD CAMELS")
        self.setWindowIcon(QIcon(str(resources.files(graphics) / "camels_icon.png")))
        layout.addWidget(self.comment_box, 0, 0)
        layout.addWidget(self.finished_box, 1, 0)
        self.setLayout(layout)
        self.adjustSize()
        self._is_finished = False
        self.finished_box.clicked.connect(self.finish_comment)
        self.setWindowFlags(self.windowFlags() & ~Qt.WindowCloseButtonHint)
        self.show()

    def finish_comment(self):
        """Sets the flag `_is_finished` to True."""
        self._is_finished = True

    def get_metadata(self):
        """Get the text of the comment."""
        return {"measurement_comments": self.comment_box.toPlainText()}

    def closeEvent(self, event):
        self.closing.emit()
        return super().closeEvent(event)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'plot_definer.ui'
##
## Created by: Qt User Interface Compiler version 6.5.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QApplication,
    QCheckBox,
    QFrame,
    QGridLayout,
    QLabel,
    QLineEdit,
    QSizePolicy,
    QSpacerItem,
    QWidget,
)

from nomad_camels.ui_widgets.variable_tool_tip_box import Variable_Box


class Ui_Plot_Definer(object):
    def setupUi(self, Plot_Definer):
        if not Plot_Definer.objectName():
            Plot_Definer.setObjectName("Plot_Definer")
        Plot_Definer.resize(470, 350)
        self.gridLayout = QGridLayout(Plot_Definer)
        self.gridLayout.setObjectName("gridLayout")
        self.lineEdit_x_axis = Variable_Box(Plot_Definer)
        self.lineEdit_x_axis.setObjectName("lineEdit_x_axis")

        self.gridLayout.addWidget(self.lineEdit_x_axis, 0, 1, 1, 3)

        self.label_5 = QLabel(Plot_Definer)
        self.label_5.setObjectName("label_5")
        font = QFont()
        font.setPointSize(9)
        font.setBold(True)
        self.label_5.setFont(font)

        self.gridLayout.addWidget(self.label_5, 0, 0, 1, 1)

        self.line = QFrame(Plot_Definer)
        self.line.setObjectName("line")
        self.line.setFrameShape(QFrame.HLine)
        self.line.setFrameShadow(QFrame.Sunken)

        self.gridLayout.addWidget(self.line, 3, 0, 1, 4)

        self.fit_definer = QLabel(Plot_Definer)
        self.fit_definer.setObjectName("fit_definer")
        font1 = QFont()
        font1.setPointSize(8)
        font1.setBold(False)
        self.fit_definer.setFont(font1)

        self.gridLayout.addWidget(self.fit_definer, 5, 3, 4, 1)

        self.checkBox_same_fit = QCheckBox(Plot_Definer)
        self.checkBox_same_fit.setObjectName("checkBox_same_fit")

        self.gridLayout.addWidget(self.checkBox_same_fit, 4, 3, 1, 1)

        self.y_axes = QWidget(Plot_Definer)
        self.y_axes.setObjectName("y_axes")

        self.gridLayout.addWidget(self.y_axes, 1, 0, 1, 4)

        self.line_2 = QFrame(Plot_Definer)
        self.line_2.setObjectName("line_2")
        self.line_2.setFrameShape(QFrame.VLine)
        self.line_2.setFrameShadow(QFrame.Sunken)

        self.gridLayout.addWidget(self.line_2, 4, 2, 5, 1)

        self.plotting_group = QWidget(Plot_Definer)
        self.plotting_group.setObjectName("plotting_group")
        self.gridLayout_2 = QGridLayout(self.plotting_group)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.gridLayout_2.setContentsMargins(0, 0, 0, 0)
        self.lineEdit_title = QLineEdit(self.plotting_group)
        self.lineEdit_title.setObjectName("lineEdit_title")

        self.gridLayout_2.addWidget(self.lineEdit_title, 1, 1, 1, 3)

        self.verticalSpacer = QSpacerItem(
            0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding
        )

        self.gridLayout_2.addItem(self.verticalSpacer, 7, 0, 1, 4)

        self.label_2 = QLabel(self.plotting_group)
        self.label_2.setObjectName("label_2")

        self.gridLayout_2.addWidget(self.label_2, 2, 0, 1, 1)

        self.checkBox_xlog = QCheckBox(self.plotting_group)
        self.checkBox_xlog.setObjectName("checkBox_xlog")

        self.gridLayout_2.addWidget(self.checkBox_xlog, 3, 0, 1, 4)

        self.checkBox_ylog = QCheckBox(self.plotting_group)
        self.checkBox_ylog.setObjectName("checkBox_ylog")

        self.gridLayout_2.addWidget(self.checkBox_ylog, 6, 0, 1, 2)

        self.label_4 = QLabel(self.plotting_group)
        self.label_4.setObjectName("label_4")

        self.gridLayout_2.addWidget(self.label_4, 4, 2, 1, 1)

        self.checkBox_ylog2 = QCheckBox(self.plotting_group)
        self.checkBox_ylog2.setObjectName("checkBox_ylog2")

        self.gridLayout_2.addWidget(self.checkBox_ylog2, 6, 2, 1, 2)

        self.label = QLabel(self.plotting_group)
        self.label.setObjectName("label")

        self.gridLayout_2.addWidget(self.label, 1, 0, 1, 1)

        self.label_3 = QLabel(self.plotting_group)
        self.label_3.setObjectName("label_3")

        self.gridLayout_2.addWidget(self.label_3, 4, 0, 1, 1)

        self.lineEdit_ylabel2 = QLineEdit(self.plotting_group)
        self.lineEdit_ylabel2.setObjectName("lineEdit_ylabel2")

        self.gridLayout_2.addWidget(self.lineEdit_ylabel2, 4, 3, 1, 1)

        self.lineEdit_xlabel = QLineEdit(self.plotting_group)
        self.lineEdit_xlabel.setObjectName("lineEdit_xlabel")

        self.gridLayout_2.addWidget(self.lineEdit_xlabel, 2, 1, 1, 3)

        self.lineEdit_ylabel = QLineEdit(self.plotting_group)
        self.lineEdit_ylabel.setObjectName("lineEdit_ylabel")

        self.gridLayout_2.addWidget(self.lineEdit_ylabel, 4, 1, 1, 1)

        self.label_6 = QLabel(self.plotting_group)
        self.label_6.setObjectName("label_6")

        self.gridLayout_2.addWidget(self.label_6, 0, 0, 1, 1)

        self.lineEdit_nPoints = QLineEdit(self.plotting_group)
        self.lineEdit_nPoints.setObjectName("lineEdit_nPoints")

        self.gridLayout_2.addWidget(self.lineEdit_nPoints, 0, 1, 1, 3)

        self.gridLayout.addWidget(self.plotting_group, 4, 0, 5, 2)

        self.retranslateUi(Plot_Definer)

        QMetaObject.connectSlotsByName(Plot_Definer)

    # setupUi

    def retranslateUi(self, Plot_Definer):
        Plot_Definer.setWindowTitle(
            QCoreApplication.translate("Plot_Definer", "Form", None)
        )
        self.label_5.setText(
            QCoreApplication.translate("Plot_Definer", "x-axis:", None)
        )
        self.fit_definer.setText(
            QCoreApplication.translate("Plot_Definer", "No y-axis selected", None)
        )
        self.checkBox_same_fit.setText(
            QCoreApplication.translate("Plot_Definer", "same fit for all ", None)
        )
        self.label_2.setText(
            QCoreApplication.translate("Plot_Definer", "x-label:", None)
        )
        self.checkBox_xlog.setText(
            QCoreApplication.translate("Plot_Definer", "x logarithmic?", None)
        )
        self.checkBox_ylog.setText(
            QCoreApplication.translate("Plot_Definer", "y logarithmic?", None)
        )
        self.label_4.setText(
            QCoreApplication.translate("Plot_Definer", "y-label 2", None)
        )
        self.checkBox_ylog2.setText(
            QCoreApplication.translate("Plot_Definer", "y2 logarithmic?", None)
        )
        self.label.setText(QCoreApplication.translate("Plot_Definer", "title:", None))
        self.label_3.setText(
            QCoreApplication.translate("Plot_Definer", "y-label:", None)
        )
        self.label_6.setText(
            QCoreApplication.translate("Plot_Definer", "# points:", None)
        )

    # retranslateUi



import pytest
from unittest.mock import patch
from pytestqt import qtbot
from unittest.mock import MagicMock
from PySide6.QtWidgets import QApplication, QDialog
from nomad_camels.MainApp_v2 import MainWindow


@pytest.fixture(scope="function")
def app(qtbot):
    """Fixture for initializing the application and the MainWindow."""
    app = QApplication.instance()
    if app is None:
        app = QApplication([])
    window = MainWindow()
    qtbot.addWidget(window)
    return window


def test_ui_initialization(app):
    """Test to check if UI components are initialized correctly."""
    assert (
        app.windowTitle()
        == "NOMAD CAMELS - Configurable Application for Measurements, Experiments and Laboratory-Systems"
    )
    assert app.pushButton_pause.icon().isNull() is False
    assert app.pushButton_resume.icon().isNull() is False
    assert app.pushButton_stop.icon().isNull() is False


def test_signal_slots(app, qtbot):
    """Test to verify that signal-slot connections are properly set."""

    # Create a mock slot
    mock_slot = MagicMock()

    # Connect the signal to the mock slot
    app.protocol_stepper_signal.connect(mock_slot)

    # Emit the signal
    app.protocol_stepper_signal.emit(42)  # Assuming the signal emits an integer

    # Assert that the mock slot was called with the expected value
    mock_slot.assert_called_once_with(42)


def test_set_user(app):
    """Test setting the user."""
    app.userdata = {"user1": {}}
    app.set_user("user1")
    assert app.active_user == "user1"

    with pytest.raises(ValueError):
        app.set_user("non_existent_user")


def test_set_sample(app):
    """Test setting the sample."""
    app.sampledata = {"sample1": {}}
    app.set_sample("sample1")
    assert app.active_sample == "sample1"

    with pytest.raises(ValueError):
        app.set_sample("non_existent_sample")


def test_show_hide_log(app, qtbot):
    """Test showing and hiding the log."""
    app.textEdit_console_output.setHidden(False)
    app.pushButton_show_log.click()
    assert app.textEdit_console_output.isHidden()
    assert app.pushButton_show_log.text() == "Show Log"

    app.pushButton_show_log.click()
    assert not app.textEdit_console_output.isHidden()
    assert app.pushButton_show_log.text() == "Hide Log"


def test_check_password_protection(app, qtbot):
    """Test checking password protection."""
    app.preferences["password_protection"] = True
    app.preferences["password_hash"] = "test_hash"

    from nomad_camels.utility.password_widgets import Password_Dialog

    with patch(
        "nomad_camels.utility.password_widgets.Password_Dialog", autospec=True
    ) as mock_dialog:
        mock_dialog.return_value.exec = MagicMock(return_value=True)
        assert app.check_password_protection() is True

    with patch(
        "nomad_camels.utility.password_widgets.Password_Dialog", autospec=True
    ) as mock_dialog:
        mock_dialog.return_value.exec = MagicMock(return_value=False)
        assert app.check_password_protection() is False

    app.preferences["password_protection"] = False
    assert app.check_password_protection() is True
